"""Tests for the MCP Tool Inspector (TC-INSPECTOR spec)."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any
from unittest.mock import AsyncMock

import pytest
from starlette.applications import Starlette
from starlette.testclient import TestClient

from apcore_mcp.inspector import create_inspector_mount

# ---------------------------------------------------------------------------
# Mock MCP Tool objects
# ---------------------------------------------------------------------------


@dataclass
class MockToolAnnotations:
    readOnlyHint: bool | None = None  # noqa: N815
    destructiveHint: bool | None = None  # noqa: N815
    idempotentHint: bool | None = None  # noqa: N815
    openWorldHint: bool | None = None  # noqa: N815
    title: str | None = None

    def model_dump(self, **kwargs: Any) -> dict[str, Any]:
        result = {}
        if self.readOnlyHint is not None:
            result["readOnlyHint"] = self.readOnlyHint
        if self.destructiveHint is not None:
            result["destructiveHint"] = self.destructiveHint
        if self.idempotentHint is not None:
            result["idempotentHint"] = self.idempotentHint
        if self.openWorldHint is not None:
            result["openWorldHint"] = self.openWorldHint
        if self.title is not None:
            result["title"] = self.title
        return result


@dataclass
class MockTool:
    name: str
    description: str
    inputSchema: dict[str, Any]  # noqa: N815
    annotations: MockToolAnnotations | None = None


# ---------------------------------------------------------------------------
# Fixtures
# ---------------------------------------------------------------------------


@pytest.fixture
def sample_tools() -> list[MockTool]:
    """Two sample MCP tools for testing."""
    return [
        MockTool(
            name="image.resize",
            description="Resize an image",
            inputSchema={
                "type": "object",
                "properties": {
                    "width": {"type": "integer"},
                    "height": {"type": "integer"},
                },
                "required": ["width", "height"],
            },
            annotations=MockToolAnnotations(readOnlyHint=False, idempotentHint=True),
        ),
        MockTool(
            name="text.echo",
            description="Echo input text",
            inputSchema={
                "type": "object",
                "properties": {"text": {"type": "string"}},
                "required": ["text"],
            },
            annotations=MockToolAnnotations(readOnlyHint=True),
        ),
    ]


@pytest.fixture
def mock_router() -> AsyncMock:
    """Mock ExecutionRouter with handle_call returning success."""
    router = AsyncMock()
    router.handle_call.return_value = (
        [{"type": "text", "text": '{"result": "ok"}'}],
        False,
    )
    return router


@pytest.fixture
def inspector_app(sample_tools: list[MockTool], mock_router: AsyncMock) -> Starlette:
    """Starlette app with inspector mounted at /inspector, allow_execute=True."""
    mount = create_inspector_mount(sample_tools, mock_router, allow_execute=True, inspector_prefix="/inspector")
    return Starlette(routes=[mount])


@pytest.fixture
def inspector_app_no_execute(sample_tools: list[MockTool], mock_router: AsyncMock) -> Starlette:
    """Starlette app with inspector mounted, allow_execute=False."""
    mount = create_inspector_mount(sample_tools, mock_router, allow_execute=False, inspector_prefix="/inspector")
    return Starlette(routes=[mount])


# ---------------------------------------------------------------------------
# TC-001: GET /inspector/ returns HTML 200 with self-contained page
# ---------------------------------------------------------------------------


class TestTC001InspectorPage:
    def test_inspector_page_returns_html(self, inspector_app: Starlette) -> None:
        client = TestClient(inspector_app)
        response = client.get("/inspector/")
        assert response.status_code == 200
        assert "text/html" in response.headers["content-type"]
        assert "MCP Tool Inspector" in response.text

    def test_inspector_page_is_self_contained(self, inspector_app: Starlette) -> None:
        client = TestClient(inspector_app)
        response = client.get("/inspector/")
        assert "<style>" in response.text
        assert "<script>" in response.text


# ---------------------------------------------------------------------------
# TC-002: Inspector disabled by default (endpoints 404 when not mounted)
# ---------------------------------------------------------------------------


class TestTC002InspectorDisabledByDefault:
    def test_no_inspector_when_not_mounted(self) -> None:
        """When inspector is not mounted, /inspector/ should 404."""
        app = Starlette(routes=[])
        client = TestClient(app, raise_server_exceptions=False)
        response = client.get("/inspector/")
        assert response.status_code == 404

    def test_no_inspector_tools_when_not_mounted(self) -> None:
        """When inspector is not mounted, /inspector/tools should 404."""
        app = Starlette(routes=[])
        client = TestClient(app, raise_server_exceptions=False)
        response = client.get("/inspector/tools")
        assert response.status_code == 404


# ---------------------------------------------------------------------------
# TC-003: GET /inspector/tools returns JSON array with correct fields
# ---------------------------------------------------------------------------


class TestTC003ListTools:
    def test_list_tools_returns_json_array(self, inspector_app: Starlette) -> None:
        client = TestClient(inspector_app)
        response = client.get("/inspector/tools")
        assert response.status_code == 200
        data = response.json()
        assert isinstance(data, list)
        assert len(data) == 2

    def test_list_tools_has_correct_fields(self, inspector_app: Starlette) -> None:
        client = TestClient(inspector_app)
        response = client.get("/inspector/tools")
        data = response.json()
        tool = data[0]
        assert "name" in tool
        assert "description" in tool
        assert tool["name"] == "image.resize"
        assert tool["description"] == "Resize an image"

    def test_list_tools_includes_annotations(self, inspector_app: Starlette) -> None:
        client = TestClient(inspector_app)
        response = client.get("/inspector/tools")
        data = response.json()
        tool = data[0]
        assert "annotations" in tool
        assert tool["annotations"]["idempotentHint"] is True


# ---------------------------------------------------------------------------
# TC-004: GET /inspector/tools/<name> returns detail + 404 for unknown
# ---------------------------------------------------------------------------


class TestTC004ToolDetail:
    def test_tool_detail_returns_full_info(self, inspector_app: Starlette) -> None:
        client = TestClient(inspector_app)
        response = client.get("/inspector/tools/image.resize")
        assert response.status_code == 200
        data = response.json()
        assert data["name"] == "image.resize"
        assert data["description"] == "Resize an image"
        assert "inputSchema" in data
        assert "properties" in data["inputSchema"]

    def test_tool_detail_includes_annotations(self, inspector_app: Starlette) -> None:
        client = TestClient(inspector_app)
        response = client.get("/inspector/tools/image.resize")
        data = response.json()
        assert "annotations" in data
        assert data["annotations"]["idempotentHint"] is True

    def test_tool_detail_404_for_unknown(self, inspector_app: Starlette) -> None:
        client = TestClient(inspector_app)
        response = client.get("/inspector/tools/nonexistent.tool")
        assert response.status_code == 404
        data = response.json()
        assert "error" in data


# ---------------------------------------------------------------------------
# TC-005: POST /inspector/tools/<name>/call executes tool
# ---------------------------------------------------------------------------


class TestTC005CallTool:
    def test_call_tool_executes(
        self,
        inspector_app: Starlette,
        mock_router: AsyncMock,
    ) -> None:
        client = TestClient(inspector_app)
        response = client.post(
            "/inspector/tools/image.resize/call",
            json={"width": 100, "height": 200},
        )
        assert response.status_code == 200
        data = response.json()
        assert "result" in data
        mock_router.handle_call.assert_called_once_with("image.resize", {"width": 100, "height": 200})

    def test_call_tool_404_for_unknown(
        self,
        inspector_app: Starlette,
    ) -> None:
        client = TestClient(inspector_app)
        response = client.post(
            "/inspector/tools/nonexistent.tool/call",
            json={},
        )
        assert response.status_code == 404

    def test_call_tool_returns_error_on_failure(
        self,
        inspector_app: Starlette,
        mock_router: AsyncMock,
    ) -> None:
        mock_router.handle_call.return_value = (
            [{"type": "text", "text": "Module not found"}],
            True,
        )
        client = TestClient(inspector_app)
        response = client.post(
            "/inspector/tools/image.resize/call",
            json={},
        )
        assert response.status_code == 500
        data = response.json()
        assert "error" in data


# ---------------------------------------------------------------------------
# TC-006: Call returns 403 when allow_execute=False
# ---------------------------------------------------------------------------


class TestTC006ExecuteDisabled:
    def test_call_returns_403_when_disabled(
        self,
        inspector_app_no_execute: Starlette,
    ) -> None:
        client = TestClient(inspector_app_no_execute)
        response = client.post(
            "/inspector/tools/image.resize/call",
            json={"width": 100, "height": 200},
        )
        assert response.status_code == 403
        data = response.json()
        assert "error" in data
        assert "disabled" in data["error"].lower() or "allow-execute" in data["error"].lower()

    def test_list_and_detail_still_work_when_execute_disabled(
        self,
        inspector_app_no_execute: Starlette,
    ) -> None:
        client = TestClient(inspector_app_no_execute)
        assert client.get("/inspector/tools").status_code == 200
        assert client.get("/inspector/tools/image.resize").status_code == 200


# ---------------------------------------------------------------------------
# TC-007: Inspector ignored for stdio (no error)
# ---------------------------------------------------------------------------


class TestTC007StdioIgnored:
    def test_explorer_flag_does_not_error_for_stdio(self) -> None:
        """When transport is stdio, explorer=True should not cause errors
        in serve() parameter validation. We test by verifying create_inspector_mount
        works and serve() validation accepts the params without transport error."""
        # The serve() function only creates the mount for HTTP transports.
        # We verify that the inspector module can be imported and mounted
        # without error, and that the serve() code path skips it for stdio.
        # Direct test: create_inspector_mount works without error
        tools = [
            MockTool(
                name="test.tool",
                description="Test",
                inputSchema={"type": "object", "properties": {}},
            )
        ]
        router = AsyncMock()
        mount = create_inspector_mount(tools, router)
        assert mount is not None


# ---------------------------------------------------------------------------
# TC-008: Custom inspector_prefix mounts at /custom/
# ---------------------------------------------------------------------------


class TestTC008CustomPrefix:
    def test_custom_prefix(
        self,
        sample_tools: list[MockTool],
        mock_router: AsyncMock,
    ) -> None:
        mount = create_inspector_mount(sample_tools, mock_router, inspector_prefix="/custom")
        app = Starlette(routes=[mount])
        client = TestClient(app)

        # Should be accessible at /custom/
        response = client.get("/custom/")
        assert response.status_code == 200
        assert "MCP Tool Inspector" in response.text

        # /custom/tools should work
        response = client.get("/custom/tools")
        assert response.status_code == 200
        data = response.json()
        assert len(data) == 2

    def test_default_prefix_not_accessible_with_custom(
        self,
        sample_tools: list[MockTool],
        mock_router: AsyncMock,
    ) -> None:
        mount = create_inspector_mount(sample_tools, mock_router, inspector_prefix="/custom")
        app = Starlette(routes=[mount])
        client = TestClient(app, raise_server_exceptions=False)

        # /inspector/ should 404 when custom prefix is used
        response = client.get("/inspector/")
        assert response.status_code == 404



"""Tests for /metrics Prometheus endpoint in TransportManager."""

from __future__ import annotations

from unittest.mock import MagicMock

from apcore_mcp.server.transport import TransportManager

PROMETHEUS_CONTENT_TYPE = "text/plain; version=0.0.4; charset=utf-8"


def _make_collector(export_text: str = "") -> MagicMock:
    """Create a mock MetricsCollector with a configurable export_prometheus() return."""
    collector = MagicMock()
    collector.export_prometheus.return_value = export_text
    return collector


# ---------------------------------------------------------------------------
# _build_metrics_response tests
# ---------------------------------------------------------------------------


class TestBuildMetricsResponse:
    """Tests for TransportManager._build_metrics_response."""

    def test_returns_404_when_no_collector(self) -> None:
        """Without a metrics collector, /metrics returns 404."""
        tm = TransportManager(metrics_collector=None)
        response = tm._build_metrics_response()
        assert response.status_code == 404

    def test_returns_200_with_collector(self) -> None:
        """With a metrics collector, /metrics returns 200."""
        collector = _make_collector("# HELP fake_counter A fake counter\n")
        tm = TransportManager(metrics_collector=collector)
        response = tm._build_metrics_response()
        assert response.status_code == 200

    def test_content_type_is_prometheus(self) -> None:
        """Response Content-Type matches Prometheus exposition format."""
        collector = _make_collector("# TYPE m counter\nm 1\n")
        tm = TransportManager(metrics_collector=collector)
        response = tm._build_metrics_response()
        assert response.media_type == PROMETHEUS_CONTENT_TYPE

    def test_body_matches_export_prometheus(self) -> None:
        """Response body is the exact output of export_prometheus()."""
        expected = '# HELP c desc\n# TYPE c counter\nc{module_id="a"} 5\n'
        collector = _make_collector(expected)
        tm = TransportManager(metrics_collector=collector)
        response = tm._build_metrics_response()
        assert bytes(response.body).decode() == expected

    def test_empty_export_returns_200(self) -> None:
        """An empty Prometheus export still returns 200 (collector is present)."""
        collector = _make_collector("")
        tm = TransportManager(metrics_collector=collector)
        response = tm._build_metrics_response()
        assert response.status_code == 200
        assert bytes(response.body).decode() == ""

    def test_export_prometheus_called_once(self) -> None:
        """export_prometheus() is called exactly once per request."""
        collector = _make_collector("data\n")
        tm = TransportManager(metrics_collector=collector)
        tm._build_metrics_response()
        collector.export_prometheus.assert_called_once()


# ---------------------------------------------------------------------------
# Constructor tests
# ---------------------------------------------------------------------------


class TestTransportManagerMetricsCollector:
    """Tests for TransportManager metrics_collector parameter."""

    def test_default_metrics_collector_is_none(self) -> None:
        """TransportManager without metrics_collector defaults to None."""
        tm = TransportManager()
        assert tm._metrics_collector is None

    def test_metrics_collector_stored(self) -> None:
        """TransportManager stores the provided metrics_collector."""
        collector = _make_collector()
        tm = TransportManager(metrics_collector=collector)
        assert tm._metrics_collector is collector



"""Unit tests for MCPServer non-blocking wrapper."""

from __future__ import annotations

import asyncio
import threading
from typing import Any
from unittest.mock import MagicMock, patch

import pytest

from apcore_mcp.server.server import MCPServer

# ---------------------------------------------------------------------------
# Stub Registry / Executor
# ---------------------------------------------------------------------------


class StubRegistry:
    """Minimal Registry stub."""

    def __init__(self) -> None:
        self._modules: dict[str, Any] = {}

    def list(self, tags: list[str] | None = None, prefix: str | None = None) -> list[str]:
        return list(self._modules.keys())

    def get_definition(self, module_id: str) -> Any:
        return self._modules.get(module_id)


class StubExecutor:
    """Minimal Executor stub with registry attribute."""

    def __init__(self) -> None:
        self.registry = StubRegistry()

    async def call_async(self, module_id: str, inputs: dict[str, Any]) -> Any:
        return {"ok": True}


# ---------------------------------------------------------------------------
# Tests for MCPServer.__init__ and properties
# ---------------------------------------------------------------------------


class TestMCPServerInit:
    """Tests for MCPServer constructor and properties."""

    def test_default_parameters(self) -> None:
        """MCPServer stores default parameters correctly."""
        registry = StubRegistry()
        server = MCPServer(registry)
        assert server._transport == "stdio"
        assert server._host == "127.0.0.1"
        assert server._port == 8000
        assert server._name == "apcore-mcp"
        assert server._version is None
        assert server._thread is None
        assert server._loop is None

    def test_custom_parameters(self) -> None:
        """MCPServer stores custom parameters correctly."""
        registry = StubRegistry()
        server = MCPServer(
            registry,
            transport="streamable-http",
            host="0.0.0.0",
            port=9000,
            name="custom-server",
            version="1.0.0",
        )
        assert server._transport == "streamable-http"
        assert server._host == "0.0.0.0"
        assert server._port == 9000
        assert server._name == "custom-server"
        assert server._version == "1.0.0"

    def test_started_and_stopped_events_initialized(self) -> None:
        """Internal threading events are properly initialized."""
        server = MCPServer(StubRegistry())
        assert isinstance(server._started, threading.Event)
        assert isinstance(server._stopped, threading.Event)
        assert not server._started.is_set()
        assert not server._stopped.is_set()


# ---------------------------------------------------------------------------
# Tests for MCPServer.address
# ---------------------------------------------------------------------------


class TestMCPServerAddress:
    """Tests for MCPServer.address property."""

    def test_stdio_address(self) -> None:
        """stdio transport returns 'stdio' as address."""
        server = MCPServer(StubRegistry(), transport="stdio")
        assert server.address == "stdio"

    def test_http_address(self) -> None:
        """HTTP transport returns formatted URL."""
        server = MCPServer(
            StubRegistry(),
            transport="streamable-http",
            host="0.0.0.0",
            port=9000,
        )
        assert server.address == "http://0.0.0.0:9000"

    def test_sse_address(self) -> None:
        """SSE transport returns formatted URL."""
        server = MCPServer(
            StubRegistry(),
            transport="sse",
            host="127.0.0.1",
            port=8080,
        )
        assert server.address == "http://127.0.0.1:8080"


# ---------------------------------------------------------------------------
# Tests for MCPServer.start / stop / wait
# ---------------------------------------------------------------------------


class TestMCPServerLifecycle:
    """Tests for MCPServer start/stop/wait lifecycle."""

    def test_start_creates_daemon_thread(self) -> None:
        """start() creates a daemon thread and waits for _started."""
        registry = StubRegistry()
        server = MCPServer(registry)

        # Mock _run to just set _started and block until _stopped
        def mock_run() -> None:
            server._started.set()
            server._stopped.wait()

        with patch.object(server, "_run", side_effect=mock_run):
            server.start()
            assert server._thread is not None
            assert server._thread.daemon is True
            assert server._started.is_set()
            server.stop()
            server._thread.join(timeout=2)

    def test_start_is_idempotent(self) -> None:
        """Calling start() twice does not create a second thread."""
        registry = StubRegistry()
        server = MCPServer(registry)

        def mock_run() -> None:
            server._started.set()
            server._stopped.wait()

        with patch.object(server, "_run", side_effect=mock_run):
            server.start()
            first_thread = server._thread
            server.start()  # Second call should be no-op
            assert server._thread is first_thread
            server.stop()
            server._thread.join(timeout=2)

    def test_wait_blocks_until_thread_finishes(self) -> None:
        """wait() blocks until the thread completes."""
        registry = StubRegistry()
        server = MCPServer(registry)

        def mock_run() -> None:
            server._started.set()
            # Immediately finish

        with patch.object(server, "_run", side_effect=mock_run):
            server.start()
            server.wait()  # Should return immediately since _run finishes
            assert not server._thread.is_alive()

    def test_wait_noop_without_start(self) -> None:
        """wait() does nothing if start() was never called."""
        server = MCPServer(StubRegistry())
        server.wait()  # Should not raise

    def test_stop_sets_stopped_event(self) -> None:
        """stop() sets the _stopped event."""
        server = MCPServer(StubRegistry())
        assert not server._stopped.is_set()
        server.stop()
        assert server._stopped.is_set()

    def test_stop_calls_loop_stop(self) -> None:
        """stop() calls loop.stop() when loop is available."""
        server = MCPServer(StubRegistry())
        mock_loop = MagicMock(spec=asyncio.AbstractEventLoop)
        server._loop = mock_loop
        server.stop()
        mock_loop.call_soon_threadsafe.assert_called_once_with(mock_loop.stop)

    def test_stop_without_loop(self) -> None:
        """stop() handles None loop gracefully."""
        server = MCPServer(StubRegistry())
        server._loop = None
        server.stop()  # Should not raise
        assert server._stopped.is_set()


# ---------------------------------------------------------------------------
# Tests for MCPServer._run
# ---------------------------------------------------------------------------


class TestMCPServerRun:
    """Tests for MCPServer._run internal method."""

    def test_run_stdio(self) -> None:
        """_run with stdio transport calls run_stdio on TransportManager."""
        registry = StubRegistry()
        server = MCPServer(registry, transport="stdio")

        with (
            patch("apcore_mcp._utils.resolve_registry", return_value=registry),
            patch("apcore_mcp._utils.resolve_executor") as mock_resolve_exec,
            patch("apcore_mcp.server.factory.MCPServerFactory") as mock_factory_cls,
            patch("apcore_mcp.server.router.ExecutionRouter"),
            patch("apcore_mcp.server.transport.TransportManager") as mock_transport_cls,
        ):
            mock_executor = MagicMock()
            mock_resolve_exec.return_value = mock_executor
            mock_tm = mock_transport_cls.return_value
            mock_factory = mock_factory_cls.return_value
            mock_mcp_server = MagicMock()
            mock_factory.create_server.return_value = mock_mcp_server
            mock_factory.build_tools.return_value = []
            mock_factory.build_init_options.return_value = MagicMock()

            # Make run_stdio a coroutine that completes immediately
            async def noop(*args: Any, **kwargs: Any) -> None:
                pass

            mock_tm.run_stdio = MagicMock(side_effect=lambda *a, **kw: noop())

            server._run()

            assert server._loop is not None
            assert server._started.is_set()
            assert server._stopped.is_set()

    def test_run_streamable_http(self) -> None:
        """_run with streamable-http transport calls run_streamable_http."""
        registry = StubRegistry()
        server = MCPServer(
            registry,
            transport="streamable-http",
            host="0.0.0.0",
            port=9000,
        )

        with (
            patch("apcore_mcp._utils.resolve_registry", return_value=registry),
            patch("apcore_mcp._utils.resolve_executor") as mock_resolve_exec,
            patch("apcore_mcp.server.factory.MCPServerFactory") as mock_factory_cls,
            patch("apcore_mcp.server.router.ExecutionRouter"),
            patch("apcore_mcp.server.transport.TransportManager") as mock_transport_cls,
        ):
            mock_resolve_exec.return_value = MagicMock()
            mock_tm = mock_transport_cls.return_value
            mock_factory = mock_factory_cls.return_value
            mock_factory.create_server.return_value = MagicMock()
            mock_factory.build_tools.return_value = []
            mock_factory.build_init_options.return_value = MagicMock()

            async def noop(*args: Any, **kwargs: Any) -> None:
                pass

            mock_tm.run_streamable_http = MagicMock(side_effect=lambda *a, **kw: noop())

            server._run()

            assert server._stopped.is_set()

    def test_run_sse(self) -> None:
        """_run with sse transport calls run_sse."""
        registry = StubRegistry()
        server = MCPServer(registry, transport="sse", host="127.0.0.1", port=8080)

        with (
            patch("apcore_mcp._utils.resolve_registry", return_value=registry),
            patch("apcore_mcp._utils.resolve_executor") as mock_resolve_exec,
            patch("apcore_mcp.server.factory.MCPServerFactory") as mock_factory_cls,
            patch("apcore_mcp.server.router.ExecutionRouter"),
            patch("apcore_mcp.server.transport.TransportManager") as mock_transport_cls,
        ):
            mock_resolve_exec.return_value = MagicMock()
            mock_tm = mock_transport_cls.return_value
            mock_factory = mock_factory_cls.return_value
            mock_factory.create_server.return_value = MagicMock()
            mock_factory.build_tools.return_value = []
            mock_factory.build_init_options.return_value = MagicMock()

            async def noop(*args: Any, **kwargs: Any) -> None:
                pass

            mock_tm.run_sse = MagicMock(side_effect=lambda *a, **kw: noop())

            server._run()

            assert server._stopped.is_set()

    def test_run_unknown_transport_raises(self) -> None:
        """_run with unknown transport raises ValueError."""
        registry = StubRegistry()
        server = MCPServer(registry, transport="unknown")

        with (
            patch("apcore_mcp._utils.resolve_registry", return_value=registry),
            patch("apcore_mcp._utils.resolve_executor") as mock_resolve_exec,
            patch("apcore_mcp.server.factory.MCPServerFactory") as mock_factory_cls,
            patch("apcore_mcp.server.router.ExecutionRouter"),
            patch("apcore_mcp.server.transport.TransportManager"),
        ):
            mock_resolve_exec.return_value = MagicMock()
            mock_factory = mock_factory_cls.return_value
            mock_factory.create_server.return_value = MagicMock()
            mock_factory.build_tools.return_value = []
            mock_factory.build_init_options.return_value = MagicMock()

            with pytest.raises(ValueError, match="Unknown transport"):
                server._run()

            # Loop should still be closed in finally block
            assert server._stopped.is_set()

    def test_run_uses_package_version_when_version_is_none(self) -> None:
        """_run uses __version__ from package when version is not specified."""
        registry = StubRegistry()
        server = MCPServer(registry, version=None)

        with (
            patch("apcore_mcp._utils.resolve_registry", return_value=registry),
            patch("apcore_mcp._utils.resolve_executor") as mock_resolve_exec,
            patch("apcore_mcp.server.factory.MCPServerFactory") as mock_factory_cls,
            patch("apcore_mcp.server.router.ExecutionRouter"),
            patch("apcore_mcp.server.transport.TransportManager") as mock_transport_cls,
        ):
            mock_resolve_exec.return_value = MagicMock()
            mock_tm = mock_transport_cls.return_value
            mock_factory = mock_factory_cls.return_value
            mock_factory.create_server.return_value = MagicMock()
            mock_factory.build_tools.return_value = []
            mock_factory.build_init_options.return_value = MagicMock()

            async def noop(*args: Any, **kwargs: Any) -> None:
                pass

            mock_tm.run_stdio = MagicMock(side_effect=lambda *a, **kw: noop())

            server._run()

            # Verify build_init_options was called with package version
            call_kwargs = mock_factory.build_init_options.call_args
            from apcore_mcp import __version__

            assert call_kwargs.kwargs["version"] == __version__ or call_kwargs[1]["version"] == __version__

    def test_run_with_executor(self) -> None:
        """_run works with an Executor (not just Registry)."""
        executor = StubExecutor()
        server = MCPServer(executor, transport="stdio")

        with (
            patch("apcore_mcp._utils.resolve_registry") as mock_resolve_reg,
            patch("apcore_mcp._utils.resolve_executor") as mock_resolve_exec,
            patch("apcore_mcp.server.factory.MCPServerFactory") as mock_factory_cls,
            patch("apcore_mcp.server.router.ExecutionRouter"),
            patch("apcore_mcp.server.transport.TransportManager") as mock_transport_cls,
        ):
            mock_resolve_reg.return_value = executor.registry
            mock_resolve_exec.return_value = executor
            mock_tm = mock_transport_cls.return_value
            mock_factory = mock_factory_cls.return_value
            mock_factory.create_server.return_value = MagicMock()
            mock_factory.build_tools.return_value = []
            mock_factory.build_init_options.return_value = MagicMock()

            async def noop(*args: Any, **kwargs: Any) -> None:
                pass

            mock_tm.run_stdio = MagicMock(side_effect=lambda *a, **kw: noop())

            server._run()

            mock_resolve_reg.assert_called_once_with(executor)
            mock_resolve_exec.assert_called_once_with(executor)

    def test_run_forwards_validate_inputs(self) -> None:
        """_run passes validate_inputs to ExecutionRouter."""
        registry = StubRegistry()
        server = MCPServer(registry, transport="stdio", validate_inputs=True)

        with (
            patch("apcore_mcp._utils.resolve_registry", return_value=registry),
            patch("apcore_mcp._utils.resolve_executor") as mock_resolve_exec,
            patch("apcore_mcp.server.factory.MCPServerFactory") as mock_factory_cls,
            patch("apcore_mcp.server.router.ExecutionRouter") as mock_router_cls,
            patch("apcore_mcp.server.transport.TransportManager") as mock_transport_cls,
        ):
            mock_executor = MagicMock()
            mock_resolve_exec.return_value = mock_executor
            mock_tm = mock_transport_cls.return_value
            mock_factory = mock_factory_cls.return_value
            mock_factory.create_server.return_value = MagicMock()
            mock_factory.build_tools.return_value = []
            mock_factory.build_init_options.return_value = MagicMock()

            async def noop(*args: Any, **kwargs: Any) -> None:
                pass

            mock_tm.run_stdio = MagicMock(side_effect=lambda *a, **kw: noop())

            server._run()

            mock_router_cls.assert_called_once_with(
                mock_executor,
                validate_inputs=True,
            )

    def test_run_closes_loop_on_error(self) -> None:
        """_run closes the event loop even if transport raises."""
        registry = StubRegistry()
        server = MCPServer(registry, transport="stdio")

        with (
            patch("apcore_mcp._utils.resolve_registry", return_value=registry),
            patch("apcore_mcp._utils.resolve_executor") as mock_resolve_exec,
            patch("apcore_mcp.server.factory.MCPServerFactory") as mock_factory_cls,
            patch("apcore_mcp.server.router.ExecutionRouter"),
            patch("apcore_mcp.server.transport.TransportManager") as mock_transport_cls,
        ):
            mock_resolve_exec.return_value = MagicMock()
            mock_tm = mock_transport_cls.return_value
            mock_factory = mock_factory_cls.return_value
            mock_factory.create_server.return_value = MagicMock()
            mock_factory.build_tools.return_value = []
            mock_factory.build_init_options.return_value = MagicMock()

            async def failing(*args: Any, **kwargs: Any) -> None:
                raise RuntimeError("Transport failed")

            mock_tm.run_stdio = MagicMock(side_effect=lambda *a, **kw: failing())

            with pytest.raises(RuntimeError, match="Transport failed"):
                server._run()

            # Loop should be closed and stopped should be set despite error
            assert server._stopped.is_set()



"""TransportManager: stdio / Streamable HTTP / SSE transport lifecycle."""

from __future__ import annotations

import logging
import time as _time
import uuid
from typing import Any, Protocol, runtime_checkable

import anyio
import uvicorn
from mcp.server.lowlevel import Server
from mcp.server.models import InitializationOptions
from mcp.server.sse import SseServerTransport
from mcp.server.stdio import stdio_server
from mcp.server.streamable_http import StreamableHTTPServerTransport
from starlette.applications import Starlette
from starlette.responses import JSONResponse, Response
from starlette.routing import Mount, Route

logger = logging.getLogger(__name__)


@runtime_checkable
class MetricsExporter(Protocol):
    """Protocol for metrics collectors that can export Prometheus text format."""

    def export_prometheus(self) -> str: ...


class TransportManager:
    """Manages MCP server transport lifecycle."""

    def __init__(self, metrics_collector: MetricsExporter | None = None) -> None:
        self._start_time = _time.monotonic()
        self._metrics_collector: MetricsExporter | None = metrics_collector
        self._module_count: int = 0

    def set_module_count(self, count: int) -> None:
        """Set the number of registered modules for health reporting."""
        self._module_count = count

    def _build_health_response(self) -> dict[str, object]:
        """Build health check response."""
        return {
            "status": "ok",
            "uptime_seconds": round(_time.monotonic() - self._start_time, 1),
            "module_count": self._module_count,
        }

    def _build_metrics_response(self) -> Response:
        """Build Prometheus metrics response.

        Returns 200 with Prometheus text if a metrics collector is configured,
        or 404 if no collector is available.
        """
        if self._metrics_collector is None:
            return Response(status_code=404)
        body = self._metrics_collector.export_prometheus()
        return Response(
            content=body,
            media_type="text/plain; version=0.0.4; charset=utf-8",
        )

    async def run_stdio(
        self,
        server: Server,
        init_options: InitializationOptions,
    ) -> None:
        """Start MCP server with stdio transport. Blocks until connection closes."""
        logger.info("Starting stdio transport")
        async with stdio_server() as (read_stream, write_stream):
            await server.run(read_stream, write_stream, init_options)

    async def run_streamable_http(
        self,
        server: Server,
        init_options: InitializationOptions,
        host: str = "127.0.0.1",
        port: int = 8000,
        extra_routes: list[Route | Mount] | None = None,
    ) -> None:
        """Start MCP server with Streamable HTTP transport."""
        self._validate_host_port(host, port)
        logger.info("Starting streamable-http transport on %s:%d", host, port)

        transport = StreamableHTTPServerTransport(
            mcp_session_id=uuid.uuid4().hex,
        )

        async with transport.connect() as (read_stream, write_stream):

            async def _health(request: Any) -> JSONResponse:
                return JSONResponse(self._build_health_response())

            async def _metrics(request: Any) -> Response:
                return self._build_metrics_response()

            routes: list[Route | Mount] = [
                Route("/health", endpoint=_health, methods=["GET"]),
                Route("/metrics", endpoint=_metrics, methods=["GET"]),
            ]
            if extra_routes:
                routes.extend(extra_routes)
            routes.append(Mount("/mcp", app=transport.handle_request))

            app = Starlette(routes=routes)

            config = uvicorn.Config(app, host=host, port=port, log_level="info")
            uv_server = uvicorn.Server(config)

            # Run both the MCP server and HTTP server concurrently
            async with anyio.create_task_group() as tg:
                tg.start_soon(server.run, read_stream, write_stream, init_options)
                tg.start_soon(uv_server.serve)

    async def run_sse(
        self,
        server: Server,
        init_options: InitializationOptions,
        host: str = "127.0.0.1",
        port: int = 8000,
        extra_routes: list[Route | Mount] | None = None,
    ) -> None:
        """Start MCP server with SSE transport (deprecated)."""
        self._validate_host_port(host, port)
        logger.info("Starting sse transport on %s:%d", host, port)
        logger.warning("SSE transport is deprecated. Use Streamable HTTP instead.")

        sse_transport = SseServerTransport("/messages/")

        async def handle_sse(request: Any) -> Response:
            async with sse_transport.connect_sse(request.scope, request.receive, request._send) as (
                read_stream,
                write_stream,
            ):
                await server.run(read_stream, write_stream, init_options)
            return Response()

        async def _health(request: Any) -> JSONResponse:
            return JSONResponse(self._build_health_response())

        async def _metrics(request: Any) -> Response:
            return self._build_metrics_response()

        routes: list[Route | Mount] = [
            Route("/health", endpoint=_health, methods=["GET"]),
            Route("/metrics", endpoint=_metrics, methods=["GET"]),
        ]
        if extra_routes:
            routes.extend(extra_routes)
        routes.extend(
            [
                Route("/sse", endpoint=handle_sse, methods=["GET"]),
                Mount("/messages/", app=sse_transport.handle_post_message),
            ]
        )

        app = Starlette(routes=routes)

        config = uvicorn.Config(app, host=host, port=port, log_level="info")
        uv_server = uvicorn.Server(config)
        await uv_server.serve()

    def _validate_host_port(self, host: str, port: int) -> None:
        """Validate host and port parameters."""
        if not host:
            raise ValueError("Host must not be empty")
        if not isinstance(port, int) or port < 1 or port > 65535:
            raise ValueError(f"Port must be between 1 and 65535, got {port}")



"""Tests for MCP extension helpers: report_progress and elicit."""

from __future__ import annotations

from typing import Any
from unittest.mock import AsyncMock

from apcore_mcp.helpers import (
    MCP_ELICIT_KEY,
    MCP_PROGRESS_KEY,
    elicit,
    report_progress,
)


class _FakeContext:
    """Minimal context stub with a data dict."""

    def __init__(self, data: dict[str, Any] | None = None) -> None:
        self.data: dict[str, Any] = data if data is not None else {}


class _NoDataContext:
    """Context stub without a data attribute."""

    pass


class TestReportProgress:
    async def test_calls_callback_with_correct_args(self) -> None:
        callback = AsyncMock()
        ctx = _FakeContext({MCP_PROGRESS_KEY: callback})

        await report_progress(ctx, 5, total=10, message="halfway there")

        callback.assert_called_once_with(5, 10, "halfway there")

    async def test_calls_callback_with_only_progress(self) -> None:
        callback = AsyncMock()
        ctx = _FakeContext({MCP_PROGRESS_KEY: callback})

        await report_progress(ctx, 3)

        callback.assert_called_once_with(3, None, None)

    async def test_noops_when_callback_absent(self) -> None:
        ctx = _FakeContext({})

        # Should not raise
        await report_progress(ctx, 1, total=10, message="test")

    async def test_noops_when_context_has_no_data_attribute(self) -> None:
        ctx = _NoDataContext()

        # Should not raise
        await report_progress(ctx, 1, total=10, message="test")


class TestElicit:
    async def test_calls_callback_and_returns_result(self) -> None:
        result = {"action": "accept", "content": {"confirmed": True}}
        callback = AsyncMock(return_value=result)
        ctx = _FakeContext({MCP_ELICIT_KEY: callback})

        response = await elicit(ctx, "Continue?", requested_schema={"type": "object"})

        callback.assert_called_once_with("Continue?", {"type": "object"})
        assert response == result

    async def test_calls_callback_without_schema(self) -> None:
        result = {"action": "decline"}
        callback = AsyncMock(return_value=result)
        ctx = _FakeContext({MCP_ELICIT_KEY: callback})

        response = await elicit(ctx, "Are you sure?")

        callback.assert_called_once_with("Are you sure?", None)
        assert response == result

    async def test_returns_none_when_callback_absent(self) -> None:
        ctx = _FakeContext({})

        response = await elicit(ctx, "Hello?")

        assert response is None

    async def test_returns_none_when_context_has_no_data_attribute(self) -> None:
        ctx = _NoDataContext()

        response = await elicit(ctx, "Hello?")

        assert response is None



"""Internal utility functions for apcore-mcp."""

from __future__ import annotations

from typing import Any


def resolve_registry(registry_or_executor: Any) -> Any:
    """Extract Registry from either a Registry or Executor instance."""
    if hasattr(registry_or_executor, "registry"):
        # It's an Executor — get its registry
        return registry_or_executor.registry
    # Assume it's a Registry
    return registry_or_executor


def resolve_executor(registry_or_executor: Any) -> Any:
    """Get or create an Executor from either a Registry or Executor instance."""
    if hasattr(registry_or_executor, "call_async"):
        # Already an Executor
        return registry_or_executor
    # It's a Registry — create a default Executor
    from apcore.executor import Executor

    return Executor(registry_or_executor)



"""ErrorMapper: apcore error hierarchy → MCP error responses."""

from __future__ import annotations

from typing import Any

from apcore_mcp.constants import ERROR_CODES


class ErrorMapper:
    """Maps apcore exceptions to MCP error response dictionaries."""

    # Error codes that should be treated as internal errors
    _INTERNAL_ERROR_CODES = {
        ERROR_CODES["CALL_DEPTH_EXCEEDED"],
        ERROR_CODES["CIRCULAR_CALL"],
        ERROR_CODES["CALL_FREQUENCY_EXCEEDED"],
    }

    # Error codes that require sanitization (hide sensitive details)
    _SANITIZED_ERROR_CODES = {
        ERROR_CODES["ACL_DENIED"],
    }

    def to_mcp_error(self, error: Exception) -> dict[str, Any]:
        """
        Convert any exception to an MCP error response dict.

        Returns:
            dict with keys:
                - is_error: True
                - error_type: str (error code or "INTERNAL_ERROR")
                - message: str (safe error message)
                - details: dict | None (optional additional context)
        """
        # Check if it's an apcore ModuleError by checking for expected attributes
        if hasattr(error, "code") and hasattr(error, "message") and hasattr(error, "details"):
            return self._handle_apcore_error(error)

        # Unknown exception - sanitize completely
        return {
            "is_error": True,
            "error_type": ERROR_CODES["INTERNAL_ERROR"],
            "message": "Internal error occurred",
            "details": None,
        }

    def _handle_apcore_error(self, error: Exception) -> dict[str, Any]:
        """Handle known apcore errors."""
        code: str = getattr(error, "code", "UNKNOWN")
        message: str = getattr(error, "message", str(error))
        raw_details: Any = getattr(error, "details", None)
        details: dict[str, Any] | None = raw_details if raw_details is not None else None

        # Convert internal errors to generic message
        if code in self._INTERNAL_ERROR_CODES:
            return {
                "is_error": True,
                "error_type": code,
                "message": "Internal error occurred",
                "details": None,
            }

        # Sanitize ACL errors to not leak caller information
        if code in self._SANITIZED_ERROR_CODES:
            return {
                "is_error": True,
                "error_type": code,
                "message": "Access denied",
                "details": None,
            }

        # Schema validation errors need special formatting
        if code == ERROR_CODES["SCHEMA_VALIDATION_ERROR"] and details is not None:
            formatted_message = self._format_validation_errors(details.get("errors", []))
            return {
                "is_error": True,
                "error_type": code,
                "message": formatted_message if formatted_message else message,
                "details": details,
            }

        # All other apcore errors: pass through message and details
        return {
            "is_error": True,
            "error_type": code,
            "message": message,
            "details": details,
        }

    def _format_validation_errors(self, errors: list[dict[str, Any]]) -> str:
        """Format SchemaValidationError field-level errors into readable message."""
        if not errors:
            return "Schema validation failed"

        # Format each error as "field: message"
        error_lines = []
        for err in errors:
            field = err.get("field", "unknown")
            msg = err.get("message", "invalid")
            error_lines.append(f"{field}: {msg}")

        return "Schema validation failed:\n" + "\n".join(f"  {line}" for line in error_lines)



"""Unit tests for RegistryListener."""

from __future__ import annotations

import logging
import threading
from typing import Any

import pytest
from mcp import types as mcp_types

from apcore_mcp.server.factory import MCPServerFactory
from apcore_mcp.server.listener import RegistryListener
from tests.conftest import ModuleAnnotations, ModuleDescriptor

# ---------------------------------------------------------------------------
# Stub Registry (with event callback support)
# ---------------------------------------------------------------------------


class StubRegistry:
    """Stub for apcore Registry with event callback support."""

    def __init__(self) -> None:
        self._callbacks: dict[str, list[Any]] = {}
        self._definitions: dict[str, ModuleDescriptor] = {}

    def on(self, event: str, callback: Any) -> None:
        self._callbacks.setdefault(event, []).append(callback)

    def get_definition(self, module_id: str) -> ModuleDescriptor | None:
        return self._definitions.get(module_id)

    def add_definition(self, descriptor: ModuleDescriptor) -> None:
        self._definitions[descriptor.module_id] = descriptor

    def trigger(self, event: str, module_id: str, module: Any = None) -> None:
        for cb in self._callbacks.get(event, []):
            cb(module_id, module)


# ---------------------------------------------------------------------------
# Shared fixtures
# ---------------------------------------------------------------------------


@pytest.fixture
def registry() -> StubRegistry:
    return StubRegistry()


@pytest.fixture
def factory() -> MCPServerFactory:
    return MCPServerFactory()


@pytest.fixture
def listener(registry: StubRegistry, factory: MCPServerFactory) -> RegistryListener:
    return RegistryListener(registry=registry, factory=factory)


def _make_descriptor(module_id: str = "test.tool", description: str = "A test tool") -> ModuleDescriptor:
    """Helper to create a simple descriptor."""
    return ModuleDescriptor(
        module_id=module_id,
        description=description,
        input_schema={
            "type": "object",
            "properties": {
                "value": {"type": "string", "description": "A value"},
            },
            "required": ["value"],
        },
        output_schema={
            "type": "object",
            "properties": {"result": {"type": "string"}},
        },
        annotations=ModuleAnnotations(readonly=True),
    )


# ---------------------------------------------------------------------------
# Tests
# ---------------------------------------------------------------------------


class TestInitialization:
    """Tests for RegistryListener initialization."""

    def test_listener_can_be_created(self, registry: StubRegistry, factory: MCPServerFactory) -> None:
        """RegistryListener can be created with a registry and factory."""
        listener = RegistryListener(registry=registry, factory=factory)
        assert listener is not None

    def test_tools_empty_initially(self, listener: RegistryListener) -> None:
        """The tools dict is empty before any events are fired."""
        assert listener.tools == {}


class TestStartStop:
    """Tests for start() and stop() lifecycle."""

    def test_start_registers_callbacks(self, listener: RegistryListener, registry: StubRegistry) -> None:
        """start() registers 'register' and 'unregister' callbacks on the registry."""
        listener.start()
        assert "register" in registry._callbacks
        assert len(registry._callbacks["register"]) == 1
        assert "unregister" in registry._callbacks
        assert len(registry._callbacks["unregister"]) == 1

    def test_start_is_idempotent(self, listener: RegistryListener, registry: StubRegistry) -> None:
        """Calling start() twice only registers callbacks once."""
        listener.start()
        listener.start()
        assert len(registry._callbacks["register"]) == 1
        assert len(registry._callbacks["unregister"]) == 1

    def test_stop_deactivates_callbacks(self, listener: RegistryListener, registry: StubRegistry) -> None:
        """After stop(), event callbacks become no-ops."""
        descriptor = _make_descriptor("stopped.tool")
        registry.add_definition(descriptor)
        listener.start()
        listener.stop()

        # Trigger register event after stop -- should be ignored
        registry.trigger("register", "stopped.tool")
        assert listener.tools == {}


class TestOnRegister:
    """Tests for _on_register callback behavior."""

    def test_on_register_adds_tool(self, listener: RegistryListener, registry: StubRegistry) -> None:
        """When a module is registered, the listener builds and adds a Tool."""
        descriptor = _make_descriptor("image.resize", "Resize an image")
        registry.add_definition(descriptor)
        listener.start()

        registry.trigger("register", "image.resize")

        tools = listener.tools
        assert "image.resize" in tools
        assert isinstance(tools["image.resize"], mcp_types.Tool)
        assert tools["image.resize"].name == "image.resize"
        assert tools["image.resize"].description == "Resize an image"

    def test_on_register_none_definition_logs_warning(
        self,
        listener: RegistryListener,
        registry: StubRegistry,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        """If get_definition returns None, logs a warning and adds no tool."""
        listener.start()

        # Do NOT add a definition -- get_definition will return None
        with caplog.at_level(logging.WARNING):
            registry.trigger("register", "ghost.module")

        assert listener.tools == {}
        assert "Definition not found for registered module: ghost.module" in caplog.text

    def test_on_register_build_tool_error_logs_warning(
        self,
        listener: RegistryListener,
        registry: StubRegistry,
        factory: MCPServerFactory,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        """If build_tool raises, logs a warning and adds no tool."""
        # Create a descriptor that will cause build_tool to fail
        bad_descriptor = ModuleDescriptor(
            module_id="bad.module",
            description="will fail",
            input_schema={
                "$defs": {},
                "properties": {"x": {"$ref": "#/$defs/Missing"}},
            },
            output_schema={},
        )
        registry.add_definition(bad_descriptor)
        listener.start()

        with caplog.at_level(logging.WARNING):
            registry.trigger("register", "bad.module")

        assert listener.tools == {}
        assert "Failed to build tool for bad.module" in caplog.text


class TestOnUnregister:
    """Tests for _on_unregister callback behavior."""

    def test_on_unregister_removes_tool(self, listener: RegistryListener, registry: StubRegistry) -> None:
        """Unregistering a module removes it from the tools dict."""
        descriptor = _make_descriptor("temp.tool")
        registry.add_definition(descriptor)
        listener.start()

        registry.trigger("register", "temp.tool")
        assert "temp.tool" in listener.tools

        registry.trigger("unregister", "temp.tool")
        assert "temp.tool" not in listener.tools

    def test_on_unregister_unknown_module_is_silent(self, listener: RegistryListener, registry: StubRegistry) -> None:
        """Unregistering an unknown module_id does not raise or add entries."""
        listener.start()

        # Should not raise
        registry.trigger("unregister", "nonexistent.tool")
        assert listener.tools == {}


class TestToolsProperty:
    """Tests for the tools property snapshot behavior."""

    def test_tools_returns_snapshot(self, listener: RegistryListener, registry: StubRegistry) -> None:
        """Modifying the returned dict does not affect internal state."""
        descriptor = _make_descriptor("snapshot.tool")
        registry.add_definition(descriptor)
        listener.start()

        registry.trigger("register", "snapshot.tool")

        # Get a snapshot and mutate it
        snapshot = listener.tools
        snapshot.pop("snapshot.tool")
        snapshot["injected"] = "bad"  # type: ignore[assignment]

        # Internal state should be unchanged
        internal = listener.tools
        assert "snapshot.tool" in internal
        assert "injected" not in internal


class TestThreadSafety:
    """Tests for concurrent register/unregister operations."""

    def test_concurrent_register_unregister(self, registry: StubRegistry, factory: MCPServerFactory) -> None:
        """Concurrent register and unregister operations do not corrupt state."""
        listener = RegistryListener(registry=registry, factory=factory)
        listener.start()

        num_modules = 100
        # Pre-add all descriptors so get_definition succeeds
        for i in range(num_modules):
            registry.add_definition(_make_descriptor(f"concurrent.tool.{i}"))

        errors: list[Exception] = []

        def register_batch(start: int, end: int) -> None:
            try:
                for i in range(start, end):
                    registry.trigger("register", f"concurrent.tool.{i}")
            except Exception as e:
                errors.append(e)

        def unregister_batch(start: int, end: int) -> None:
            try:
                for i in range(start, end):
                    registry.trigger("unregister", f"concurrent.tool.{i}")
            except Exception as e:
                errors.append(e)

        threads = []
        # Register first half while unregistering concurrently
        t1 = threading.Thread(target=register_batch, args=(0, num_modules))
        t2 = threading.Thread(target=unregister_batch, args=(0, num_modules // 2))
        threads.extend([t1, t2])

        for t in threads:
            t.start()
        for t in threads:
            t.join(timeout=10)

        # No exceptions should have been raised
        assert errors == [], f"Concurrent operations raised errors: {errors}"

        # The tools dict should be in a consistent state (a valid dict, no corruption)
        tools = listener.tools
        assert isinstance(tools, dict)
        # All values should be Tool instances
        for tool in tools.values():
            assert isinstance(tool, mcp_types.Tool)



"""Non-blocking MCP server wrapper for framework integrations."""

from __future__ import annotations

import asyncio
import logging
import threading
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from apcore_mcp.server.transport import MetricsExporter

logger = logging.getLogger(__name__)


class MCPServer:
    """Non-blocking MCP server.

    Usage:
        server = MCPServer(registry, transport="streamable-http", port=8000)
        server.start()
        print(f"Server running at {server.address}")
        server.wait()  # blocks until shutdown
    """

    def __init__(
        self,
        registry_or_executor: object,
        *,
        transport: str = "stdio",
        host: str = "127.0.0.1",
        port: int = 8000,
        name: str = "apcore-mcp",
        version: str | None = None,
        validate_inputs: bool = False,
        metrics_collector: MetricsExporter | None = None,
        tags: list[str] | None = None,
        prefix: str | None = None,
    ) -> None:
        self._registry_or_executor = registry_or_executor
        self._transport = transport.lower()
        self._host = host
        self._port = port
        self._name = name
        self._version = version
        self._validate_inputs = validate_inputs
        self._metrics_collector = metrics_collector
        self._tags = tags
        self._prefix = prefix
        self._thread: threading.Thread | None = None
        self._loop: asyncio.AbstractEventLoop | None = None
        self._started = threading.Event()
        self._stopped = threading.Event()

    @property
    def address(self) -> str:
        """Server address (available after start)."""
        if self._transport == "stdio":
            return "stdio"
        return f"http://{self._host}:{self._port}"

    def start(self) -> None:
        """Start the server in a background thread (non-blocking)."""
        if self._thread is not None:
            return
        self._thread = threading.Thread(target=self._run, daemon=True)
        self._thread.start()
        self._started.wait(timeout=10)

    def wait(self) -> None:
        """Block until the server stops."""
        if self._thread is not None:
            self._thread.join()

    def stop(self) -> None:
        """Gracefully stop the server."""
        if self._loop is not None:
            self._loop.call_soon_threadsafe(self._loop.stop)
        self._stopped.set()

    def _run(self) -> None:
        """Internal: run the server event loop."""
        from importlib.metadata import version as _pkg_version

        from apcore_mcp._utils import resolve_executor, resolve_registry

        __version__ = _pkg_version("apcore-mcp")
        from apcore_mcp.server.factory import MCPServerFactory
        from apcore_mcp.server.router import ExecutionRouter
        from apcore_mcp.server.transport import TransportManager

        registry = resolve_registry(self._registry_or_executor)
        executor = resolve_executor(self._registry_or_executor)
        version = self._version or __version__

        factory = MCPServerFactory()
        server = factory.create_server(name=self._name, version=version)
        tools = factory.build_tools(registry, tags=self._tags, prefix=self._prefix)
        router = ExecutionRouter(executor, validate_inputs=self._validate_inputs)
        factory.register_handlers(server, tools, router)
        factory.register_resource_handlers(server, registry)
        init_options = factory.build_init_options(
            server,
            name=self._name,
            version=version,
        )

        transport_manager = TransportManager(metrics_collector=self._metrics_collector)
        transport_manager.set_module_count(len(tools))

        self._loop = asyncio.new_event_loop()
        self._started.set()

        try:
            if self._transport == "stdio":
                self._loop.run_until_complete(
                    transport_manager.run_stdio(server, init_options),
                )
            elif self._transport == "streamable-http":
                self._loop.run_until_complete(
                    transport_manager.run_streamable_http(
                        server,
                        init_options,
                        host=self._host,
                        port=self._port,
                    ),
                )
            elif self._transport == "sse":
                self._loop.run_until_complete(
                    transport_manager.run_sse(
                        server,
                        init_options,
                        host=self._host,
                        port=self._port,
                    ),
                )
            else:
                msg = f"Unknown transport: {self._transport}"
                raise ValueError(msg)
        finally:
            self._loop.close()
            self._stopped.set()



"""Tests for ErrorMapper: apcore errors → MCP error response dicts."""

from __future__ import annotations

from typing import Any

import pytest

from apcore_mcp.adapters.errors import ErrorMapper


# Stub error classes that mimic apcore error hierarchy for testing
# This avoids hard dependency on apcore in unit tests
class ModuleError(Exception):
    """Base error for all apcore framework errors."""

    def __init__(
        self,
        code: str,
        message: str,
        details: dict[str, Any] | None = None,
        cause: Exception | None = None,
        trace_id: str | None = None,
    ) -> None:
        super().__init__(message)
        self.code = code
        self.message = message
        self.details: dict[str, Any] = details or {}
        self.cause = cause
        self.trace_id = trace_id


class ModuleNotFoundError(ModuleError):
    """Raised when a module cannot be found."""

    def __init__(self, module_id: str, **kwargs: Any) -> None:
        super().__init__(
            code="MODULE_NOT_FOUND",
            message=f"Module not found: {module_id}",
            details={"module_id": module_id},
            **kwargs,
        )


class SchemaValidationError(ModuleError):
    """Raised when schema validation fails."""

    def __init__(
        self,
        message: str = "Schema validation failed",
        errors: list[dict[str, Any]] | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            code="SCHEMA_VALIDATION_ERROR",
            message=message,
            details={"errors": errors or []},
            **kwargs,
        )


class ACLDeniedError(ModuleError):
    """Raised when ACL denies access."""

    def __init__(self, caller_id: str | None, target_id: str, **kwargs: Any) -> None:
        super().__init__(
            code="ACL_DENIED",
            message=f"Access denied: {caller_id} -> {target_id}",
            details={"caller_id": caller_id, "target_id": target_id},
            **kwargs,
        )


class ModuleTimeoutError(ModuleError):
    """Raised when module execution exceeds timeout."""

    def __init__(self, module_id: str, timeout_ms: int, **kwargs: Any) -> None:
        super().__init__(
            code="MODULE_TIMEOUT",
            message=f"Module {module_id} timed out after {timeout_ms}ms",
            details={"module_id": module_id, "timeout_ms": timeout_ms},
            **kwargs,
        )


class InvalidInputError(ModuleError):
    """Raised for invalid input."""

    def __init__(self, message: str = "Invalid input", **kwargs: Any) -> None:
        super().__init__(code="GENERAL_INVALID_INPUT", message=message, **kwargs)


class CallDepthExceededError(ModuleError):
    """Raised when call chain exceeds maximum depth."""

    def __init__(self, depth: int, max_depth: int, call_chain: list[str], **kwargs: Any) -> None:
        super().__init__(
            code="CALL_DEPTH_EXCEEDED",
            message=f"Call depth {depth} exceeds maximum {max_depth}",
            details={"depth": depth, "max_depth": max_depth, "call_chain": call_chain},
            **kwargs,
        )


class CircularCallError(ModuleError):
    """Raised when a circular call is detected."""

    def __init__(self, module_id: str, call_chain: list[str], **kwargs: Any) -> None:
        super().__init__(
            code="CIRCULAR_CALL",
            message=f"Circular call detected for module {module_id}",
            details={"module_id": module_id, "call_chain": call_chain},
            **kwargs,
        )


class CallFrequencyExceededError(ModuleError):
    """Raised when a module is called too many times."""

    def __init__(
        self,
        module_id: str,
        count: int,
        max_repeat: int,
        call_chain: list[str],
        **kwargs: Any,
    ) -> None:
        super().__init__(
            code="CALL_FREQUENCY_EXCEEDED",
            message=f"Module {module_id} called {count} times, max is {max_repeat}",
            details={
                "module_id": module_id,
                "count": count,
                "max_repeat": max_repeat,
                "call_chain": call_chain,
            },
            **kwargs,
        )


# Test cases
class TestErrorMapper:
    """Test suite for ErrorMapper."""

    @pytest.fixture
    def mapper(self) -> ErrorMapper:
        """Create an ErrorMapper instance."""
        return ErrorMapper()

    def test_module_not_found(self, mapper: ErrorMapper) -> None:
        """ModuleNotFoundError contains module_id in message and error_type."""
        error = ModuleNotFoundError("image.resize")
        result = mapper.to_mcp_error(error)

        assert result["is_error"] is True
        assert result["error_type"] == "MODULE_NOT_FOUND"
        assert "image.resize" in result["message"]

    def test_schema_validation_error(self, mapper: ErrorMapper) -> None:
        """SchemaValidationError includes field details in message."""
        errors = [
            {"field": "name", "message": "required"},
            {"field": "age", "message": "must be positive"},
        ]
        error = SchemaValidationError("Validation failed", errors=errors)
        result = mapper.to_mcp_error(error)

        assert result["is_error"] is True
        assert result["error_type"] == "SCHEMA_VALIDATION_ERROR"
        assert "name" in result["message"]
        assert "age" in result["message"]

    def test_acl_denied(self, mapper: ErrorMapper) -> None:
        """ACLDeniedError says access denied but does NOT leak caller_id."""
        error = ACLDeniedError("user1", "admin.delete")
        result = mapper.to_mcp_error(error)

        assert result["is_error"] is True
        assert result["error_type"] == "ACL_DENIED"
        assert "access denied" in result["message"].lower()
        # Should NOT contain sensitive caller_id
        assert "user1" not in result["message"]

    def test_module_timeout(self, mapper: ErrorMapper) -> None:
        """ModuleTimeoutError mentions timeout."""
        error = ModuleTimeoutError("slow.module", 5000)
        result = mapper.to_mcp_error(error)

        assert result["is_error"] is True
        assert result["error_type"] == "MODULE_TIMEOUT"
        assert "timeout" in result["message"].lower() or "timed out" in result["message"].lower()

    def test_invalid_input(self, mapper: ErrorMapper) -> None:
        """InvalidInputError preserves message."""
        error = InvalidInputError("missing field X")
        result = mapper.to_mcp_error(error)

        assert result["is_error"] is True
        assert result["error_type"] == "GENERAL_INVALID_INPUT"
        assert "missing field X" in result["message"]

    def test_call_depth_exceeded(self, mapper: ErrorMapper) -> None:
        """CallDepthExceededError becomes internal error."""
        error = CallDepthExceededError(10, 10, ["a", "b"])
        result = mapper.to_mcp_error(error)

        assert result["is_error"] is True
        assert result["error_type"] == "CALL_DEPTH_EXCEEDED"
        assert "internal" in result["message"].lower()

    def test_circular_call(self, mapper: ErrorMapper) -> None:
        """CircularCallError becomes internal error."""
        error = CircularCallError("a", ["a", "b", "a"])
        result = mapper.to_mcp_error(error)

        assert result["is_error"] is True
        assert result["error_type"] == "CIRCULAR_CALL"
        assert "internal" in result["message"].lower()

    def test_call_frequency_exceeded(self, mapper: ErrorMapper) -> None:
        """CallFrequencyExceededError becomes internal error."""
        error = CallFrequencyExceededError("a", 5, 3, ["a"])
        result = mapper.to_mcp_error(error)

        assert result["is_error"] is True
        assert result["error_type"] == "CALL_FREQUENCY_EXCEEDED"
        assert "internal" in result["message"].lower()

    def test_unexpected_exception(self, mapper: ErrorMapper) -> None:
        """Unexpected exceptions become generic internal error with NO stack trace."""
        error = ValueError("oops")
        result = mapper.to_mcp_error(error)

        assert result["is_error"] is True
        assert result["error_type"] == "INTERNAL_ERROR"
        assert "internal error" in result["message"].lower()
        # Should NOT leak the original error message
        assert "oops" not in result["message"]

    def test_all_errors_set_is_error(self, mapper: ErrorMapper) -> None:
        """Every error type returns is_error=True."""
        errors = [
            ModuleNotFoundError("test"),
            SchemaValidationError("test"),
            ACLDeniedError("user", "admin"),
            ModuleTimeoutError("test", 100),
            InvalidInputError("test"),
            CallDepthExceededError(1, 1, []),
            CircularCallError("test", []),
            CallFrequencyExceededError("test", 5, 3, []),
            ValueError("unexpected"),
        ]

        for error in errors:
            result = mapper.to_mcp_error(error)
            assert result["is_error"] is True, f"{type(error).__name__} should set is_error=True"

    def test_sanitize_no_stack_trace(self, mapper: ErrorMapper) -> None:
        """Unexpected exceptions don't leak traceback info."""
        error = RuntimeError("internal details that should not be exposed")
        result = mapper.to_mcp_error(error)

        # Should not contain any of the internal error details
        assert "internal details" not in result["message"]
        assert "RuntimeError" not in result["message"]
        # Should be generic message
        assert result["message"] == "Internal error occurred"



"""Shared test fixtures for apcore-mcp tests."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

import pytest

# ---------------------------------------------------------------------------
# Lightweight stubs for apcore types used in unit tests.
# These mirror the real apcore API surface without importing apcore,
# so that adapter/converter unit tests have zero external dependencies.
# ---------------------------------------------------------------------------


@dataclass(frozen=True)
class ModuleAnnotations:
    """Stub for apcore.module.ModuleAnnotations."""

    readonly: bool = False
    destructive: bool = False
    idempotent: bool = False
    requires_approval: bool = False
    open_world: bool = True


@dataclass
class ModuleExample:
    """Stub for apcore.module.ModuleExample."""

    title: str
    inputs: dict[str, Any] = field(default_factory=dict)
    output: dict[str, Any] = field(default_factory=dict)
    description: str | None = None


@dataclass
class ModuleDescriptor:
    """Stub for apcore.registry.types.ModuleDescriptor."""

    module_id: str
    description: str
    input_schema: dict[str, Any]
    output_schema: dict[str, Any]
    name: str | None = None
    documentation: str | None = None
    version: str = "1.0.0"
    tags: list[str] = field(default_factory=list)
    annotations: ModuleAnnotations | None = None
    examples: list[ModuleExample] = field(default_factory=list)
    metadata: dict[str, Any] = field(default_factory=dict)


# ---------------------------------------------------------------------------
# Fixtures: reusable ModuleDescriptor instances for tests
# ---------------------------------------------------------------------------


@pytest.fixture
def simple_descriptor() -> ModuleDescriptor:
    """A simple module with flat input schema."""
    return ModuleDescriptor(
        module_id="image.resize",
        name="Image Resize",
        description="Resize an image to the specified dimensions",
        input_schema={
            "type": "object",
            "properties": {
                "width": {"type": "integer", "description": "Target width in pixels"},
                "height": {"type": "integer", "description": "Target height in pixels"},
                "image_path": {
                    "type": "string",
                    "description": "Path to the image file",
                },
            },
            "required": ["width", "height", "image_path"],
        },
        output_schema={
            "type": "object",
            "properties": {
                "output_path": {"type": "string"},
                "original_size": {"type": "array", "items": {"type": "integer"}},
                "new_size": {"type": "array", "items": {"type": "integer"}},
            },
            "required": ["output_path"],
        },
        tags=["image", "transform"],
        annotations=ModuleAnnotations(idempotent=True),
    )


@pytest.fixture
def empty_schema_descriptor() -> ModuleDescriptor:
    """A module with empty input/output schemas."""
    return ModuleDescriptor(
        module_id="system.ping",
        description="Health check ping",
        input_schema={},
        output_schema={},
        annotations=ModuleAnnotations(readonly=True, idempotent=True),
    )


@pytest.fixture
def nested_schema_descriptor() -> ModuleDescriptor:
    """A module with nested/complex input schema including $defs."""
    return ModuleDescriptor(
        module_id="workflow.execute",
        name="Workflow Execute",
        description="Execute a multi-step workflow",
        input_schema={
            "type": "object",
            "$defs": {
                "Step": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "params": {"type": "object"},
                    },
                    "required": ["name"],
                }
            },
            "properties": {
                "workflow_name": {"type": "string"},
                "steps": {
                    "type": "array",
                    "items": {"$ref": "#/$defs/Step"},
                },
                "dry_run": {"type": "boolean", "default": False},
            },
            "required": ["workflow_name", "steps"],
        },
        output_schema={
            "type": "object",
            "properties": {
                "results": {"type": "array", "items": {"type": "object"}},
                "success": {"type": "boolean"},
            },
        },
        annotations=ModuleAnnotations(destructive=True, requires_approval=True, open_world=False),
    )


@pytest.fixture
def destructive_descriptor() -> ModuleDescriptor:
    """A module marked as destructive."""
    return ModuleDescriptor(
        module_id="file.delete",
        description="Delete a file from disk",
        input_schema={
            "type": "object",
            "properties": {
                "path": {"type": "string"},
                "recursive": {"type": "boolean", "default": False},
            },
            "required": ["path"],
        },
        output_schema={
            "type": "object",
            "properties": {"deleted": {"type": "boolean"}},
        },
        annotations=ModuleAnnotations(destructive=True, requires_approval=True, open_world=False),
    )


@pytest.fixture
def no_annotations_descriptor() -> ModuleDescriptor:
    """A module with no annotations (None)."""
    return ModuleDescriptor(
        module_id="text.echo",
        description="Echo the input text",
        input_schema={
            "type": "object",
            "properties": {"text": {"type": "string"}},
            "required": ["text"],
        },
        output_schema={
            "type": "object",
            "properties": {"text": {"type": "string"}},
        },
        annotations=None,
    )


@pytest.fixture
def all_types_descriptor() -> ModuleDescriptor:
    """A module using all JSON Schema types in its input."""
    return ModuleDescriptor(
        module_id="test.all_types",
        description="Test module with all JSON Schema types",
        input_schema={
            "type": "object",
            "properties": {
                "str_field": {"type": "string"},
                "int_field": {"type": "integer"},
                "num_field": {"type": "number"},
                "bool_field": {"type": "boolean"},
                "null_field": {"type": "null"},
                "arr_field": {"type": "array", "items": {"type": "string"}},
                "obj_field": {
                    "type": "object",
                    "properties": {"nested": {"type": "string"}},
                },
            },
        },
        output_schema={"type": "object"},
    )



"""Unit tests for AnnotationMapper."""

from __future__ import annotations

import pytest

from apcore_mcp.adapters.annotations import AnnotationMapper
from tests.conftest import ModuleAnnotations


class TestAnnotationMapper:
    """Test suite for AnnotationMapper."""

    @pytest.fixture
    def mapper(self) -> AnnotationMapper:
        """Create an AnnotationMapper instance for testing."""
        return AnnotationMapper()

    def test_readonly_annotation(self, mapper: AnnotationMapper) -> None:
        """Test readonly annotation maps to read_only_hint=True."""
        annotations = ModuleAnnotations(readonly=True)
        result = mapper.to_mcp_annotations(annotations)

        assert result == {
            "read_only_hint": True,
            "destructive_hint": False,
            "idempotent_hint": False,
            "open_world_hint": True,
            "title": None,
        }

    def test_destructive_annotation(self, mapper: AnnotationMapper) -> None:
        """Test destructive annotation maps to destructive_hint=True."""
        annotations = ModuleAnnotations(destructive=True)
        result = mapper.to_mcp_annotations(annotations)

        assert result == {
            "read_only_hint": False,
            "destructive_hint": True,
            "idempotent_hint": False,
            "open_world_hint": True,
            "title": None,
        }

    def test_idempotent_annotation(self, mapper: AnnotationMapper) -> None:
        """Test idempotent annotation maps to idempotent_hint=True."""
        annotations = ModuleAnnotations(idempotent=True)
        result = mapper.to_mcp_annotations(annotations)

        assert result == {
            "read_only_hint": False,
            "destructive_hint": False,
            "idempotent_hint": True,
            "open_world_hint": True,
            "title": None,
        }

    def test_open_world_false(self, mapper: AnnotationMapper) -> None:
        """Test open_world=False maps to open_world_hint=False."""
        annotations = ModuleAnnotations(open_world=False)
        result = mapper.to_mcp_annotations(annotations)

        assert result == {
            "read_only_hint": False,
            "destructive_hint": False,
            "idempotent_hint": False,
            "open_world_hint": False,
            "title": None,
        }

    def test_all_defaults(self, mapper: AnnotationMapper) -> None:
        """Test default ModuleAnnotations maps to default MCP annotations."""
        annotations = ModuleAnnotations()
        result = mapper.to_mcp_annotations(annotations)

        assert result == {
            "read_only_hint": False,
            "destructive_hint": False,
            "idempotent_hint": False,
            "open_world_hint": True,
            "title": None,
        }

    def test_none_annotations(self, mapper: AnnotationMapper) -> None:
        """Test None annotations uses default values."""
        result = mapper.to_mcp_annotations(None)

        assert result == {
            "read_only_hint": False,
            "destructive_hint": False,
            "idempotent_hint": False,
            "open_world_hint": True,
            "title": None,
        }

    def test_combined_annotations(self, mapper: AnnotationMapper) -> None:
        """Test multiple annotations combine correctly."""
        annotations = ModuleAnnotations(
            destructive=True,
            requires_approval=True,
            open_world=False,
        )
        result = mapper.to_mcp_annotations(annotations)

        assert result == {
            "read_only_hint": False,
            "destructive_hint": True,
            "idempotent_hint": False,
            "open_world_hint": False,
            "title": None,
        }

    def test_has_requires_approval_true(self, mapper: AnnotationMapper) -> None:
        """Test has_requires_approval returns True when requires_approval=True."""
        annotations = ModuleAnnotations(requires_approval=True)
        assert mapper.has_requires_approval(annotations) is True

    def test_has_requires_approval_false(self, mapper: AnnotationMapper) -> None:
        """Test has_requires_approval returns False for default annotations."""
        annotations = ModuleAnnotations()
        assert mapper.has_requires_approval(annotations) is False

    def test_has_requires_approval_none(self, mapper: AnnotationMapper) -> None:
        """Test has_requires_approval returns False for None annotations."""
        assert mapper.has_requires_approval(None) is False

    def test_description_suffix_with_annotations(self, mapper: AnnotationMapper) -> None:
        """Test description suffix embeds only non-default annotations."""
        annotations = ModuleAnnotations(
            destructive=True,
            readonly=False,
            idempotent=True,
            requires_approval=True,
            open_world=False,
        )
        result = mapper.to_description_suffix(annotations)

        # Expected format: newlines + bracketed annotation list
        assert result.startswith("\n\n[Annotations: ")
        assert result.endswith("]")
        assert "destructive=true" in result.lower()
        assert "idempotent=true" in result.lower()
        assert "requires_approval=true" in result.lower()
        assert "open_world=false" in result.lower()
        # readonly=False is the default, so it should NOT appear
        assert "readonly" not in result.lower()

    def test_description_suffix_empty(self, mapper: AnnotationMapper) -> None:
        """Test description suffix returns empty string for None annotations."""
        result = mapper.to_description_suffix(None)
        assert result == ""

    def test_description_suffix_defaults(self, mapper: AnnotationMapper) -> None:
        """Test description suffix with all-default annotations returns empty string."""
        annotations = ModuleAnnotations()
        result = mapper.to_description_suffix(annotations)

        # All-default annotations should return empty string
        assert result == ""



"""Security tests for apcore-mcp.

TC-SEC-001 through TC-SEC-006: ACL leak prevention, stack trace sanitization,
malformed input handling, oversized input handling, call chain info hiding,
and default host binding.
"""

from __future__ import annotations

import inspect
from typing import Any

import pytest

import apcore_mcp
from apcore_mcp.adapters.errors import ErrorMapper
from apcore_mcp.server.router import ExecutionRouter
from apcore_mcp.server.transport import TransportManager

# ---------------------------------------------------------------------------
# Stub error classes (same pattern as tests/server/test_router.py)
# ---------------------------------------------------------------------------


class StubModuleError(Exception):
    """Base stub for apcore ModuleError."""

    def __init__(
        self,
        code: str,
        message: str,
        details: dict[str, Any] | None = None,
    ) -> None:
        super().__init__(message)
        self.code = code
        self.message = message
        self.details: dict[str, Any] = details or {}


class ACLDeniedStubError(StubModuleError):
    """Stub for apcore ACLDeniedError."""

    def __init__(self, caller_id: str | None, target_id: str) -> None:
        super().__init__(
            code="ACL_DENIED",
            message=f"Access denied: {caller_id} -> {target_id}",
            details={"caller_id": caller_id, "target_id": target_id},
        )


class CallDepthExceededStubError(StubModuleError):
    """Stub for apcore CallDepthExceededError."""

    def __init__(self, depth: int, max_depth: int, call_chain: list[str]) -> None:
        super().__init__(
            code="CALL_DEPTH_EXCEEDED",
            message=f"Call depth {depth} exceeds maximum {max_depth}",
            details={"depth": depth, "max_depth": max_depth, "call_chain": call_chain},
        )


class CircularCallStubError(StubModuleError):
    """Stub for apcore CircularCallError."""

    def __init__(self, module_id: str, call_chain: list[str]) -> None:
        super().__init__(
            code="CIRCULAR_CALL",
            message=f"Circular call detected for module {module_id}",
            details={"module_id": module_id, "call_chain": call_chain},
        )


class CallFrequencyExceededStubError(StubModuleError):
    """Stub for apcore CallFrequencyExceededError."""

    def __init__(self, module_id: str, count: int, max_repeat: int, call_chain: list[str]) -> None:
        super().__init__(
            code="CALL_FREQUENCY_EXCEEDED",
            message=f"Module {module_id} called {count} times, max is {max_repeat}",
            details={
                "module_id": module_id,
                "count": count,
                "max_repeat": max_repeat,
                "call_chain": call_chain,
            },
        )


# ---------------------------------------------------------------------------
# Stub executor
# ---------------------------------------------------------------------------


class StubExecutor:
    """Stub executor that raises a configured error or returns a result."""

    def __init__(
        self,
        results: dict[str, Any] | None = None,
        errors: dict[str, Exception] | None = None,
        default_error: Exception | None = None,
    ) -> None:
        self._results = results or {}
        self._errors = errors or {}
        self._default_error = default_error

    async def call_async(self, module_id: str, inputs: dict[str, Any] | None = None) -> Any:
        if module_id in self._errors:
            raise self._errors[module_id]
        if self._default_error:
            raise self._default_error
        return self._results.get(module_id, {"ok": True})


# ---------------------------------------------------------------------------
# TC-SEC-001: ACL denied calls do not leak caller identity
# ---------------------------------------------------------------------------


class TestACLDeniedNoLeak:
    """TC-SEC-001: ACL denied responses must not expose caller identity."""

    async def test_acl_denied_does_not_leak_caller_id(self) -> None:
        """Response to ACL denied must not contain caller_id and must say 'Access denied'."""
        error = ACLDeniedStubError(
            caller_id="admin_user_42",
            target_id="secret.module",
        )
        executor = StubExecutor(
            errors={"secret.module": error},
        )
        router = ExecutionRouter(executor)

        content, is_error = await router.handle_call("secret.module", {})

        assert is_error is True
        response_text = content[0]["text"]
        assert "admin_user_42" not in response_text, "Response must not contain caller identity"
        assert response_text == "Access denied", f"Expected exactly 'Access denied', got '{response_text}'"


# ---------------------------------------------------------------------------
# TC-SEC-002: Unexpected exceptions do not leak stack traces
# ---------------------------------------------------------------------------


class TestNoStackTraceLeak:
    """TC-SEC-002: unexpected exceptions must not expose sensitive details."""

    @pytest.mark.parametrize(
        "exception,forbidden_substrings",
        [
            (
                RuntimeError("Connection refused to postgres://admin:secret@db:5432/prod"),
                ["postgres", "secret"],
            ),
            (
                FileNotFoundError("/etc/shadow"),
                ["/etc/shadow"],
            ),
            (
                PermissionError("Cannot access /root/.ssh/id_rsa"),
                ["/root/.ssh"],
            ),
            (
                ValueError("Invalid token: eyJhbGciOiJIUzI1NiJ9.eyJ0ZXN0IjoiZGF0YSJ9"),
                ["eyJhbG"],
            ),
        ],
        ids=[
            "postgres_connection_string",
            "shadow_file_path",
            "ssh_private_key_path",
            "jwt_token",
        ],
    )
    async def test_sensitive_exception_sanitized(
        self,
        exception: Exception,
        forbidden_substrings: list[str],
    ) -> None:
        """Unexpected exceptions must be mapped to 'Internal error occurred'."""
        executor = StubExecutor(default_error=exception)
        router = ExecutionRouter(executor)

        content, is_error = await router.handle_call("any.module", {})

        assert is_error is True
        response_text = content[0]["text"]
        assert response_text == "Internal error occurred", f"Expected 'Internal error occurred', got '{response_text}'"
        for substring in forbidden_substrings:
            assert substring not in response_text, f"Response must not contain '{substring}'"


# ---------------------------------------------------------------------------
# TC-SEC-003: Malformed input handled safely
# ---------------------------------------------------------------------------


class TestMalformedInput:
    """TC-SEC-003: malformed/malicious inputs must not cause unhandled exceptions."""

    @pytest.mark.parametrize(
        "malicious_input",
        [
            # XSS payload
            {"field": "<script>alert('xss')</script>"},
            # SQL injection
            {"field": "'; DROP TABLE users; --"},
            # Template injection
            {"field": "{{7*7}}"},
            # Null bytes
            {"field": "hello\x00world"},
            # Unicode edge cases
            {"field": "\ud800"},
            # Deeply nested dict
            {"a": {"b": {"c": {"d": {"e": {"f": "deep"}}}}}},
            # Array with mixed types
            {"items": [1, "two", None, True, {"nested": "obj"}]},
            # Empty string key
            {"": "empty_key"},
            # Very long key name
            {"k" * 10000: "long_key"},
        ],
        ids=[
            "xss_payload",
            "sql_injection",
            "template_injection",
            "null_bytes",
            "unicode_edge_case",
            "deeply_nested",
            "mixed_array",
            "empty_key",
            "long_key_name",
        ],
    )
    async def test_malformed_input_handled_safely(self, malicious_input: dict[str, Any]) -> None:
        """Malicious inputs produce either success or error, never unhandled exception."""
        executor = StubExecutor(results={"test.module": {"ok": True}})
        router = ExecutionRouter(executor)

        # Must not raise -- either success or graceful error
        content, is_error = await router.handle_call("test.module", malicious_input)

        assert isinstance(content, list), "content must be a list"
        assert len(content) >= 1, "content must have at least one element"
        assert isinstance(is_error, bool), "is_error must be a boolean"
        assert content[0]["type"] == "text", "content type must be 'text'"


# ---------------------------------------------------------------------------
# TC-SEC-004: Oversized input handled gracefully
# ---------------------------------------------------------------------------


class TestOversizedInput:
    """TC-SEC-004: very large inputs must not cause memory exhaustion crashes."""

    async def test_oversized_10mb_string_handled(self) -> None:
        """A 10MB string value in arguments does not crash the router."""
        large_value = "A" * (10 * 1024 * 1024)  # 10MB
        arguments = {"data": large_value}

        executor = StubExecutor(results={"test.module": {"ok": True}})
        router = ExecutionRouter(executor)

        # Must not crash -- either success or error is acceptable
        content, is_error = await router.handle_call("test.module", arguments)

        assert isinstance(content, list), "content must be a list"
        assert len(content) >= 1, "content must have at least one element"
        assert isinstance(is_error, bool), "is_error must be a boolean"

    async def test_oversized_many_keys_handled(self) -> None:
        """Arguments with 10,000 keys do not crash the router."""
        arguments = {f"key_{i}": f"value_{i}" for i in range(10_000)}

        executor = StubExecutor(results={"test.module": {"ok": True}})
        router = ExecutionRouter(executor)

        content, is_error = await router.handle_call("test.module", arguments)

        assert isinstance(content, list)
        assert isinstance(is_error, bool)


# ---------------------------------------------------------------------------
# TC-SEC-005: Error mapper never exposes call chain details
# ---------------------------------------------------------------------------


class TestCallChainNotExposed:
    """TC-SEC-005: error mapper must never expose call chain details."""

    @pytest.mark.parametrize(
        "error,forbidden_substrings",
        [
            (
                CallDepthExceededStubError(
                    depth=10,
                    max_depth=5,
                    call_chain=["module.a", "module.b", "module.c"],
                ),
                ["module.a", "module.b", "module.c"],
            ),
            (
                CircularCallStubError(
                    module_id="module.x",
                    call_chain=["module.x", "module.y", "module.x"],
                ),
                ["module.x", "module.y"],
            ),
            (
                CallFrequencyExceededStubError(
                    module_id="spam",
                    count=100,
                    max_repeat=10,
                    call_chain=["spam", "helper.a"],
                ),
                ["spam", "helper.a"],
            ),
        ],
        ids=[
            "call_depth_exceeded",
            "circular_call",
            "call_frequency_exceeded",
        ],
    )
    async def test_call_chain_not_in_response(
        self,
        error: Exception,
        forbidden_substrings: list[str],
    ) -> None:
        """Call chain module names must not appear in the error response."""
        executor = StubExecutor(default_error=error)
        router = ExecutionRouter(executor)

        content, is_error = await router.handle_call("any.module", {})

        assert is_error is True
        response_text = content[0]["text"]
        assert response_text == "Internal error occurred", f"Expected 'Internal error occurred', got '{response_text}'"
        for substring in forbidden_substrings:
            assert substring not in response_text, f"Response must not contain call chain module '{substring}'"

    def test_error_mapper_directly_hides_call_chain(self) -> None:
        """ErrorMapper.to_mcp_error strips call chain from all internal error types."""
        mapper = ErrorMapper()

        errors = [
            CallDepthExceededStubError(10, 5, ["module.a", "module.b"]),
            CircularCallStubError("module.x", ["module.x", "module.y"]),
            CallFrequencyExceededStubError("spam", 100, 10, ["spam", "helper.a"]),
        ]

        for error in errors:
            result = mapper.to_mcp_error(error)
            assert (
                result["message"] == "Internal error occurred"
            ), f"ErrorMapper must return 'Internal error occurred' for {type(error).__name__}"
            assert result["details"] is None, f"ErrorMapper must not expose details for {type(error).__name__}"


# ---------------------------------------------------------------------------
# TC-SEC-006: Default HTTP host is localhost only
# ---------------------------------------------------------------------------


class TestDefaultHostLocalhost:
    """TC-SEC-006: all default HTTP host parameters must be 127.0.0.1."""

    def test_serve_default_host_is_localhost(self) -> None:
        """serve() function default host parameter is 127.0.0.1."""
        sig = inspect.signature(apcore_mcp.serve)
        host_param = sig.parameters["host"]
        assert (
            host_param.default == "127.0.0.1"
        ), f"serve() default host is '{host_param.default}', expected '127.0.0.1'"

    def test_transport_streamable_http_default_host_is_localhost(self) -> None:
        """TransportManager.run_streamable_http default host is 127.0.0.1."""
        sig = inspect.signature(TransportManager.run_streamable_http)
        host_param = sig.parameters["host"]
        assert (
            host_param.default == "127.0.0.1"
        ), f"run_streamable_http default host is '{host_param.default}', expected '127.0.0.1'"

    def test_transport_sse_default_host_is_localhost(self) -> None:
        """TransportManager.run_sse default host is 127.0.0.1."""
        sig = inspect.signature(TransportManager.run_sse)
        host_param = sig.parameters["host"]
        assert (
            host_param.default == "127.0.0.1"
        ), f"run_sse default host is '{host_param.default}', expected '127.0.0.1'"



"""RegistryListener: dynamic tool registration on registry changes."""

from __future__ import annotations

import logging
import threading
from typing import Any

from mcp import types as mcp_types

from apcore_mcp.constants import REGISTRY_EVENTS
from apcore_mcp.server.factory import MCPServerFactory

logger = logging.getLogger(__name__)


class RegistryListener:
    """Listens for Registry changes and updates MCP tool list."""

    def __init__(
        self,
        registry: Any,
        factory: MCPServerFactory,
    ) -> None:
        """Initialize the listener.

        Args:
            registry: apcore Registry to listen to.
            factory: MCPServerFactory for building Tool objects from new modules.
        """
        self._registry = registry
        self._factory = factory
        self._tools: dict[str, mcp_types.Tool] = {}
        self._lock = threading.Lock()
        self._active = False

    @property
    def tools(self) -> dict[str, mcp_types.Tool]:
        """Return a snapshot of currently registered tools. Thread-safe."""
        with self._lock:
            return dict(self._tools)

    def start(self) -> None:
        """Start listening for Registry events.

        Registers callbacks via registry.on("register", ...) and
        registry.on("unregister", ...).
        Safe to call multiple times (idempotent).
        """
        if self._active:
            return
        self._active = True
        self._registry.on(REGISTRY_EVENTS["REGISTER"], self._on_register)
        self._registry.on(REGISTRY_EVENTS["UNREGISTER"], self._on_unregister)

    def stop(self) -> None:
        """Stop listening for Registry events.

        Sets internal flag that causes callbacks to no-op.
        (apcore Registry does not support callback removal)
        """
        self._active = False

    def _on_register(self, module_id: str, module: Any = None) -> None:
        """Callback for Registry 'register' event.

        Steps:
        1. Check if listener is active (no-op if stopped)
        2. Call registry.get_definition(module_id)
        3. If None (race condition), log warning and return
        4. Call factory.build_tool(descriptor)
        5. Add to internal _tools dict (thread-safe via lock)
        6. Log info: "Tool registered: {module_id}"
        """
        if not self._active:
            return
        descriptor = self._registry.get_definition(module_id)
        if descriptor is None:
            logger.warning("Definition not found for registered module: %s", module_id)
            return
        try:
            tool = self._factory.build_tool(descriptor)
            with self._lock:
                self._tools[module_id] = tool
            logger.info("Tool registered: %s", module_id)
        except Exception as e:
            logger.warning("Failed to build tool for %s: %s", module_id, e)

    def _on_unregister(self, module_id: str, module: Any = None) -> None:
        """Callback for Registry 'unregister' event.

        Steps:
        1. Check if listener is active (no-op if stopped)
        2. Remove module_id from _tools dict (thread-safe)
        3. If module_id not in dict, silently ignore
        4. Log info: "Tool unregistered: {module_id}"
        """
        if not self._active:
            return
        with self._lock:
            removed = self._tools.pop(module_id, None)
        if removed is not None:
            logger.info("Tool unregistered: %s", module_id)



"""End-to-end tests for apcore-mcp.

These tests validate the full system behavior from the perspective of an
external consumer, using real components throughout the stack. Only the
Executor and Registry are stubbed since they represent the apcore boundary.

Tests that require real subprocess/network (TC-E2E-001, TC-E2E-002,
TC-E2E-003) are skipped because they need real apcore modules.
"""

from __future__ import annotations

import pytest
from mcp import types as mcp_types

from apcore_mcp import to_openai_tools
from apcore_mcp.server.factory import MCPServerFactory
from tests.conftest import ModuleAnnotations, ModuleDescriptor

# ---------------------------------------------------------------------------
# Stubs
# ---------------------------------------------------------------------------


class StubRegistry:
    """Stub for apcore Registry."""

    def __init__(self, descriptors: list[ModuleDescriptor] | None = None) -> None:
        self._descriptors: dict[str, ModuleDescriptor] = {d.module_id: d for d in (descriptors or [])}

    def list(self, tags: list[str] | None = None, prefix: str | None = None) -> list[str]:
        ids = list(self._descriptors.keys())
        if prefix is not None:
            ids = [mid for mid in ids if mid.startswith(prefix)]
        if tags is not None:
            tag_set = set(tags)
            ids = [mid for mid in ids if tag_set.issubset(set(self._descriptors[mid].tags))]
        return sorted(ids)

    def get_definition(self, module_id: str) -> ModuleDescriptor | None:
        return self._descriptors.get(module_id)


# ---------------------------------------------------------------------------
# TC-E2E-004: Multi-module registry with mixed annotations
# ---------------------------------------------------------------------------


class TestMultiModuleMixedAnnotations:
    """TC-E2E-004: Build MCP tools from a multi-module registry with varied annotations.

    Three modules with different annotation profiles:
    - reader.get: readonly=True, idempotent=True
    - writer.delete: destructive=True, requires_approval=True
    - worker.process: annotations=None (defaults)

    Validates that the full pipeline (MCPServerFactory with real SchemaConverter
    and AnnotationMapper) produces correct MCP ToolAnnotations for each module.
    """

    @pytest.fixture
    def reader_desc(self) -> ModuleDescriptor:
        return ModuleDescriptor(
            module_id="reader.get",
            description="Read data from a source",
            input_schema={
                "type": "object",
                "properties": {
                    "source": {
                        "type": "string",
                        "description": "Data source identifier",
                    },
                    "limit": {"type": "integer", "description": "Max items to return"},
                },
                "required": ["source"],
            },
            output_schema={
                "type": "object",
                "properties": {"items": {"type": "array", "items": {"type": "object"}}},
            },
            annotations=ModuleAnnotations(readonly=True, idempotent=True),
        )

    @pytest.fixture
    def writer_desc(self) -> ModuleDescriptor:
        return ModuleDescriptor(
            module_id="writer.delete",
            description="Delete records from the database",
            input_schema={
                "type": "object",
                "properties": {
                    "record_id": {"type": "string", "description": "ID of the record"},
                    "force": {"type": "boolean", "description": "Force delete"},
                },
                "required": ["record_id"],
            },
            output_schema={
                "type": "object",
                "properties": {"deleted": {"type": "boolean"}},
            },
            annotations=ModuleAnnotations(destructive=True, requires_approval=True, open_world=False),
        )

    @pytest.fixture
    def worker_desc(self) -> ModuleDescriptor:
        return ModuleDescriptor(
            module_id="worker.process",
            description="Process a batch of items",
            input_schema={
                "type": "object",
                "properties": {
                    "batch_id": {"type": "string"},
                    "concurrency": {"type": "integer"},
                },
                "required": ["batch_id"],
            },
            output_schema={
                "type": "object",
                "properties": {"processed": {"type": "integer"}},
            },
            annotations=None,
        )

    @pytest.fixture
    def registry(
        self,
        reader_desc: ModuleDescriptor,
        writer_desc: ModuleDescriptor,
        worker_desc: ModuleDescriptor,
    ) -> StubRegistry:
        return StubRegistry([reader_desc, writer_desc, worker_desc])

    @pytest.fixture
    def tools(self, registry: StubRegistry) -> list[mcp_types.Tool]:
        factory = MCPServerFactory()
        return factory.build_tools(registry)

    @pytest.fixture
    def tool_by_name(self, tools: list[mcp_types.Tool]) -> dict[str, mcp_types.Tool]:
        return {t.name: t for t in tools}

    def test_three_tools_built(self, tools: list[mcp_types.Tool]) -> None:
        """All three modules produce MCP Tools."""
        assert len(tools) == 3

    def test_tool_names(self, tool_by_name: dict[str, mcp_types.Tool]) -> None:
        """Tool names match module IDs."""
        assert set(tool_by_name.keys()) == {
            "reader.get",
            "writer.delete",
            "worker.process",
        }

    # -- reader.get annotations --

    def test_reader_readonly_hint(self, tool_by_name: dict[str, mcp_types.Tool]) -> None:
        """reader.get has readOnlyHint=True."""
        tool = tool_by_name["reader.get"]
        assert tool.annotations is not None
        assert tool.annotations.readOnlyHint is True

    def test_reader_idempotent_hint(self, tool_by_name: dict[str, mcp_types.Tool]) -> None:
        """reader.get has idempotentHint=True."""
        tool = tool_by_name["reader.get"]
        assert tool.annotations is not None
        assert tool.annotations.idempotentHint is True

    def test_reader_destructive_hint_false(self, tool_by_name: dict[str, mcp_types.Tool]) -> None:
        """reader.get has destructiveHint=False (default)."""
        tool = tool_by_name["reader.get"]
        assert tool.annotations is not None
        assert tool.annotations.destructiveHint is False

    def test_reader_open_world_hint_default(self, tool_by_name: dict[str, mcp_types.Tool]) -> None:
        """reader.get has openWorldHint=True (default)."""
        tool = tool_by_name["reader.get"]
        assert tool.annotations is not None
        assert tool.annotations.openWorldHint is True

    # -- writer.delete annotations --

    def test_writer_destructive_hint(self, tool_by_name: dict[str, mcp_types.Tool]) -> None:
        """writer.delete has destructiveHint=True."""
        tool = tool_by_name["writer.delete"]
        assert tool.annotations is not None
        assert tool.annotations.destructiveHint is True

    def test_writer_open_world_hint_false(self, tool_by_name: dict[str, mcp_types.Tool]) -> None:
        """writer.delete has openWorldHint=False."""
        tool = tool_by_name["writer.delete"]
        assert tool.annotations is not None
        assert tool.annotations.openWorldHint is False

    def test_writer_readonly_hint_false(self, tool_by_name: dict[str, mcp_types.Tool]) -> None:
        """writer.delete has readOnlyHint=False (default)."""
        tool = tool_by_name["writer.delete"]
        assert tool.annotations is not None
        assert tool.annotations.readOnlyHint is False

    def test_writer_idempotent_hint_false(self, tool_by_name: dict[str, mcp_types.Tool]) -> None:
        """writer.delete has idempotentHint=False (default)."""
        tool = tool_by_name["writer.delete"]
        assert tool.annotations is not None
        assert tool.annotations.idempotentHint is False

    # -- worker.process annotations (None -> defaults) --

    def test_worker_default_readonly_hint(self, tool_by_name: dict[str, mcp_types.Tool]) -> None:
        """worker.process (annotations=None) has readOnlyHint=False."""
        tool = tool_by_name["worker.process"]
        assert tool.annotations is not None
        assert tool.annotations.readOnlyHint is False

    def test_worker_default_destructive_hint(self, tool_by_name: dict[str, mcp_types.Tool]) -> None:
        """worker.process (annotations=None) has destructiveHint=False."""
        tool = tool_by_name["worker.process"]
        assert tool.annotations is not None
        assert tool.annotations.destructiveHint is False

    def test_worker_default_idempotent_hint(self, tool_by_name: dict[str, mcp_types.Tool]) -> None:
        """worker.process (annotations=None) has idempotentHint=False."""
        tool = tool_by_name["worker.process"]
        assert tool.annotations is not None
        assert tool.annotations.idempotentHint is False

    def test_worker_default_open_world_hint(self, tool_by_name: dict[str, mcp_types.Tool]) -> None:
        """worker.process (annotations=None) has openWorldHint=True."""
        tool = tool_by_name["worker.process"]
        assert tool.annotations is not None
        assert tool.annotations.openWorldHint is True

    # -- Schema validation for all tools --

    def test_all_tools_have_valid_input_schema(self, tools: list[mcp_types.Tool]) -> None:
        """All tools have inputSchema with type: 'object'."""
        for tool in tools:
            assert tool.inputSchema["type"] == "object"
            assert "properties" in tool.inputSchema

    def test_all_tools_have_descriptions(self, tools: list[mcp_types.Tool]) -> None:
        """All tools have non-empty descriptions."""
        for tool in tools:
            assert tool.description is not None
            assert len(tool.description) > 0


# ---------------------------------------------------------------------------
# TC-E2E-008: Verify server wires with zero modules
# ---------------------------------------------------------------------------


class TestZeroModules:
    """TC-E2E-008: Server components work correctly with an empty registry.

    Validates that both to_openai_tools() and MCPServerFactory.build_tools()
    gracefully handle a registry with no modules.
    """

    @pytest.fixture
    def empty_registry(self) -> StubRegistry:
        return StubRegistry([])

    def test_openai_tools_returns_empty_list(self, empty_registry: StubRegistry) -> None:
        """to_openai_tools() returns an empty list for an empty registry."""
        tools = to_openai_tools(empty_registry)
        assert isinstance(tools, list)
        assert tools == []

    def test_build_tools_returns_empty_list(self, empty_registry: StubRegistry) -> None:
        """MCPServerFactory.build_tools() returns an empty list for an empty registry."""
        factory = MCPServerFactory()
        tools = factory.build_tools(empty_registry)
        assert isinstance(tools, list)
        assert tools == []

    def test_openai_tools_return_type(self, empty_registry: StubRegistry) -> None:
        """to_openai_tools() always returns a list, not None."""
        result = to_openai_tools(empty_registry)
        assert result is not None
        assert isinstance(result, list)

    def test_build_tools_return_type(self, empty_registry: StubRegistry) -> None:
        """MCPServerFactory.build_tools() always returns a list, not None."""
        factory = MCPServerFactory()
        result = factory.build_tools(empty_registry)
        assert result is not None
        assert isinstance(result, list)



"""Starlette route handlers for the MCP Tool Inspector."""

from __future__ import annotations

import json
import logging
from typing import Any

from starlette.requests import Request
from starlette.responses import HTMLResponse, JSONResponse, Response
from starlette.routing import Route

from apcore_mcp.inspector.html import _INSPECTOR_HTML

logger = logging.getLogger(__name__)


def _make_serializable(obj: Any) -> Any:
    """Convert Pydantic models and other non-JSON-serializable objects to dicts."""
    if hasattr(obj, "model_dump"):
        return obj.model_dump(exclude_none=True)
    if hasattr(obj, "dict"):
        return obj.dict(exclude_none=True)
    return obj


def _tool_summary(tool: Any) -> dict[str, Any]:
    """Return a summary dict for a tool (used in the list endpoint)."""
    result: dict[str, Any] = {
        "name": tool.name,
        "description": tool.description or "",
    }
    annotations = _make_serializable(getattr(tool, "annotations", None))
    if annotations:
        result["annotations"] = annotations
    return result


def _tool_detail(tool: Any) -> dict[str, Any]:
    """Return a full detail dict for a tool (used in the detail endpoint)."""
    result: dict[str, Any] = {
        "name": tool.name,
        "description": tool.description or "",
        "inputSchema": _make_serializable(tool.inputSchema),
    }
    annotations = _make_serializable(getattr(tool, "annotations", None))
    if annotations:
        result["annotations"] = annotations
    return result


def build_inspector_routes(
    tools: list[Any],
    router: Any,
    *,
    allow_execute: bool = False,
) -> list[Route]:
    """Build Starlette routes for the MCP Tool Inspector.

    Args:
        tools: List of MCP Tool objects.
        router: An ExecutionRouter with async handle_call(name, arguments).
        allow_execute: Whether to allow tool execution via the call endpoint.

    Returns:
        List of Starlette Route objects to be mounted under the inspector prefix.
    """
    tools_by_name: dict[str, Any] = {t.name: t for t in tools}

    async def inspector_page(request: Request) -> HTMLResponse:
        return HTMLResponse(_INSPECTOR_HTML)

    async def list_tools(request: Request) -> JSONResponse:
        return JSONResponse([_tool_summary(t) for t in tools])

    async def tool_detail(request: Request) -> Response:
        name = request.path_params["name"]
        tool = tools_by_name.get(name)
        if tool is None:
            return JSONResponse({"error": f"Tool not found: {name}"}, status_code=404)
        return JSONResponse(_tool_detail(tool))

    async def call_tool(request: Request) -> Response:
        if not allow_execute:
            return JSONResponse(
                {"error": "Tool execution is disabled. Launch with --allow-execute to enable."},
                status_code=403,
            )
        name = request.path_params["name"]
        tool = tools_by_name.get(name)
        if tool is None:
            return JSONResponse({"error": f"Tool not found: {name}"}, status_code=404)

        try:
            body = await request.json()
        except Exception:
            body = {}

        try:
            content, is_error = await router.handle_call(name, body)
            # Extract text from content list
            texts = [item.get("text", "") for item in content if item.get("type") == "text"]
            # Try to parse as JSON for a cleaner response
            if len(texts) == 1:
                try:
                    result = json.loads(texts[0])
                except (json.JSONDecodeError, TypeError):
                    result = texts[0]
            else:
                result = texts

            if is_error:
                return JSONResponse({"error": result}, status_code=500)
            return JSONResponse({"result": result})
        except Exception as exc:
            logger.error("Inspector call_tool error for %s: %s", name, exc)
            return JSONResponse({"error": str(exc)}, status_code=500)

    return [
        Route("/", endpoint=inspector_page, methods=["GET"]),
        Route("/tools", endpoint=list_tools, methods=["GET"]),
        Route("/tools/{name:path}/call", endpoint=call_tool, methods=["POST"]),
        Route("/tools/{name:path}", endpoint=tool_detail, methods=["GET"]),
    ]



"""Tests for the apcore-mcp public API: serve() and to_openai_tools()."""

from __future__ import annotations

from unittest.mock import AsyncMock, MagicMock, patch

import pytest

import apcore_mcp
from apcore_mcp import serve, to_openai_tools
from apcore_mcp._utils import resolve_executor, resolve_registry
from tests.conftest import ModuleAnnotations, ModuleDescriptor

# ---------------------------------------------------------------------------
# Stub Registry / Executor for public API tests
# ---------------------------------------------------------------------------


class StubRegistry:
    """Minimal Registry stub with list() and get_definition()."""

    def __init__(self, descriptors: list[ModuleDescriptor] | None = None):
        self._descriptors = {d.module_id: d for d in (descriptors or [])}

    def list(self, tags=None, prefix=None):
        ids = list(self._descriptors.keys())
        if prefix is not None:
            ids = [mid for mid in ids if mid.startswith(prefix)]
        if tags is not None:
            tag_set = set(tags)
            ids = [mid for mid in ids if tag_set.issubset(set(self._descriptors[mid].tags))]
        return sorted(ids)

    def get_definition(self, module_id):
        return self._descriptors.get(module_id)


class StubExecutor:
    """Minimal Executor stub with call_async() and registry attribute."""

    def __init__(self, registry: StubRegistry):
        self.registry = registry

    async def call_async(self, _module_id: str, _inputs: dict | None = None) -> dict:
        return {"ok": True}


# ---------------------------------------------------------------------------
# Fixtures
# ---------------------------------------------------------------------------


@pytest.fixture
def sample_descriptors() -> list[ModuleDescriptor]:
    return [
        ModuleDescriptor(
            module_id="image.resize",
            description="Resize an image",
            input_schema={
                "type": "object",
                "properties": {
                    "width": {"type": "integer"},
                    "height": {"type": "integer"},
                },
                "required": ["width", "height"],
            },
            output_schema={"type": "object"},
            tags=["image"],
            annotations=ModuleAnnotations(idempotent=True),
        ),
        ModuleDescriptor(
            module_id="text.echo",
            description="Echo text",
            input_schema={
                "type": "object",
                "properties": {"text": {"type": "string"}},
                "required": ["text"],
            },
            output_schema={"type": "object"},
            tags=["text"],
        ),
    ]


@pytest.fixture
def registry(sample_descriptors) -> StubRegistry:
    return StubRegistry(sample_descriptors)


@pytest.fixture
def executor(registry) -> StubExecutor:
    return StubExecutor(registry)


# ===========================================================================
# Tests for resolve_registry
# ===========================================================================


class TestResolveRegistry:
    """Tests for resolve_registry helper."""

    def test_returns_registry_from_executor(self, registry, executor):
        """When given an executor, extracts its .registry attribute."""
        result = resolve_registry(executor)
        assert result is registry

    def test_returns_registry_directly(self, registry):
        """When given a registry (no .registry attr), returns it as-is."""
        result = resolve_registry(registry)
        assert result is registry


# ===========================================================================
# Tests for resolve_executor
# ===========================================================================


class TestResolveExecutor:
    """Tests for resolve_executor helper."""

    def test_returns_executor_directly(self, executor):
        """When given an executor (has .call_async), returns it as-is."""
        result = resolve_executor(executor)
        assert result is executor

    def test_creates_executor_from_registry(self, registry):
        """When given a registry, imports and creates an Executor."""
        import sys

        mock_executor_cls = MagicMock()
        mock_executor_instance = MagicMock()
        mock_executor_cls.return_value = mock_executor_instance

        fake_apcore_executor = MagicMock()
        fake_apcore_executor.Executor = mock_executor_cls
        sys.modules["apcore"] = MagicMock()
        sys.modules["apcore.executor"] = fake_apcore_executor

        try:
            result = resolve_executor(registry)
            mock_executor_cls.assert_called_once_with(registry)
            assert result is mock_executor_instance
        finally:
            del sys.modules["apcore.executor"]
            del sys.modules["apcore"]


# ===========================================================================
# Tests for to_openai_tools
# ===========================================================================


class TestToOpenaiTools:
    """Tests for to_openai_tools() public API."""

    def test_basic_conversion_with_registry(self, registry):
        """Converts registry modules to OpenAI tool format."""
        tools = to_openai_tools(registry)
        assert isinstance(tools, list)
        assert len(tools) == 2

        # Check structure of first tool
        tool = tools[0]
        assert tool["type"] == "function"
        assert "function" in tool
        assert "name" in tool["function"]
        assert "description" in tool["function"]
        assert "parameters" in tool["function"]

    def test_conversion_with_executor(self, executor):
        """Accepts an Executor and extracts its registry for conversion."""
        tools = to_openai_tools(executor)
        assert len(tools) == 2

    def test_tool_names_are_normalized(self, registry):
        """Module IDs with dots are converted to dashes."""
        tools = to_openai_tools(registry)
        names = {t["function"]["name"] for t in tools}
        assert "image-resize" in names
        assert "text-echo" in names

    def test_embed_annotations(self, registry):
        """embed_annotations=True appends annotation info to descriptions."""
        tools = to_openai_tools(registry, embed_annotations=True)
        # Find the tool with annotations
        image_tool = next(t for t in tools if t["function"]["name"] == "image-resize")
        assert "[Annotations:" in image_tool["function"]["description"]

    def test_strict_mode(self, registry):
        """strict=True adds strict: true to function definitions."""
        tools = to_openai_tools(registry, strict=True)
        for tool in tools:
            assert tool["function"]["strict"] is True

    def test_filter_by_tags(self, registry):
        """tags parameter filters modules by tag."""
        tools = to_openai_tools(registry, tags=["image"])
        assert len(tools) == 1
        assert tools[0]["function"]["name"] == "image-resize"

    def test_filter_by_prefix(self, registry):
        """prefix parameter filters modules by ID prefix."""
        tools = to_openai_tools(registry, prefix="text")
        assert len(tools) == 1
        assert tools[0]["function"]["name"] == "text-echo"

    def test_empty_registry(self):
        """Empty registry returns empty list."""
        empty_reg = StubRegistry([])
        tools = to_openai_tools(empty_reg)
        assert tools == []


# ===========================================================================
# Tests for serve()
# ===========================================================================


class TestServe:
    """Tests for serve() public API."""

    def test_serve_stdio_wires_components(self, registry):
        """serve() with stdio transport wires factory, router, transport correctly."""
        with (
            patch("apcore_mcp.TransportManager") as mock_tm_cls,
            patch("apcore_mcp.MCPServerFactory") as mock_factory_cls,
            patch("apcore_mcp.ExecutionRouter"),
        ):
            # Set up factory mock
            mock_factory = mock_factory_cls.return_value
            mock_server = MagicMock()
            mock_factory.create_server.return_value = mock_server
            mock_factory.build_tools.return_value = []
            mock_factory.build_init_options.return_value = MagicMock()

            # Set up transport mock
            mock_tm = mock_tm_cls.return_value
            mock_tm.run_stdio = AsyncMock()

            serve(registry, transport="stdio", name="test-server", version="1.0.0")

            mock_factory.create_server.assert_called_once_with(name="test-server", version="1.0.0")
            mock_factory.build_tools.assert_called_once_with(registry, tags=None, prefix=None)
            mock_factory.register_handlers.assert_called_once()
            mock_factory.build_init_options.assert_called_once()

    def test_serve_streamable_http(self, registry):
        """serve() with streamable-http transport passes host and port."""
        with (
            patch("apcore_mcp.TransportManager") as mock_tm_cls,
            patch("apcore_mcp.MCPServerFactory") as mock_factory_cls,
            patch("apcore_mcp.ExecutionRouter"),
        ):
            mock_factory = mock_factory_cls.return_value
            mock_factory.create_server.return_value = MagicMock()
            mock_factory.build_tools.return_value = []
            mock_factory.build_init_options.return_value = MagicMock()

            mock_tm = mock_tm_cls.return_value
            mock_tm.run_streamable_http = AsyncMock()

            serve(registry, transport="streamable-http", host="0.0.0.0", port=9000)

            mock_tm.run_streamable_http.assert_called_once()
            call_kwargs = mock_tm.run_streamable_http.call_args
            assert call_kwargs.kwargs["host"] == "0.0.0.0"
            assert call_kwargs.kwargs["port"] == 9000

    def test_serve_sse(self, registry):
        """serve() with sse transport passes host and port."""
        with (
            patch("apcore_mcp.TransportManager") as mock_tm_cls,
            patch("apcore_mcp.MCPServerFactory") as mock_factory_cls,
            patch("apcore_mcp.ExecutionRouter"),
        ):
            mock_factory = mock_factory_cls.return_value
            mock_factory.create_server.return_value = MagicMock()
            mock_factory.build_tools.return_value = []
            mock_factory.build_init_options.return_value = MagicMock()

            mock_tm = mock_tm_cls.return_value
            mock_tm.run_sse = AsyncMock()

            serve(registry, transport="sse", host="localhost", port=8080)

            mock_tm.run_sse.assert_called_once()
            call_kwargs = mock_tm.run_sse.call_args
            assert call_kwargs.kwargs["host"] == "localhost"
            assert call_kwargs.kwargs["port"] == 8080

    def test_serve_unknown_transport(self, registry):
        """serve() with unknown transport raises ValueError."""
        with (
            patch("apcore_mcp.TransportManager"),
            patch("apcore_mcp.MCPServerFactory") as mock_factory_cls,
            patch("apcore_mcp.ExecutionRouter"),
        ):
            mock_factory = mock_factory_cls.return_value
            mock_factory.create_server.return_value = MagicMock()
            mock_factory.build_tools.return_value = []
            mock_factory.build_init_options.return_value = MagicMock()

            with pytest.raises(ValueError, match="Unknown transport.*'websocket'"):
                serve(registry, transport="websocket")

    def test_serve_default_version(self, registry):
        """serve() uses __version__ when version is not specified."""
        with (
            patch("apcore_mcp.TransportManager") as mock_tm_cls,
            patch("apcore_mcp.MCPServerFactory") as mock_factory_cls,
            patch("apcore_mcp.ExecutionRouter"),
        ):
            mock_factory = mock_factory_cls.return_value
            mock_factory.create_server.return_value = MagicMock()
            mock_factory.build_tools.return_value = []
            mock_factory.build_init_options.return_value = MagicMock()

            mock_tm = mock_tm_cls.return_value
            mock_tm.run_stdio = AsyncMock()

            serve(registry)

            call_kwargs = mock_factory.create_server.call_args
            assert call_kwargs.kwargs["version"] == apcore_mcp.__version__

    def test_serve_with_executor(self, executor):
        """serve() accepts an Executor and extracts its registry."""
        with (
            patch("apcore_mcp.TransportManager") as mock_tm_cls,
            patch("apcore_mcp.MCPServerFactory") as mock_factory_cls,
            patch("apcore_mcp.ExecutionRouter") as mock_router_cls,
        ):
            mock_factory = mock_factory_cls.return_value
            mock_factory.create_server.return_value = MagicMock()
            mock_factory.build_tools.return_value = []
            mock_factory.build_init_options.return_value = MagicMock()

            mock_tm = mock_tm_cls.return_value
            mock_tm.run_stdio = AsyncMock()

            serve(executor)

            # Router should receive the executor with validate_inputs
            mock_router_cls.assert_called_once_with(executor, validate_inputs=False)
            # Factory should receive the extracted registry
            mock_factory.build_tools.assert_called_once_with(executor.registry, tags=None, prefix=None)



"""Adapters: schema conversion, annotation mapping, error mapping, ID normalization."""

from apcore_mcp.adapters.annotations import AnnotationMapper
from apcore_mcp.adapters.errors import ErrorMapper
from apcore_mcp.adapters.id_normalizer import ModuleIDNormalizer

__all__ = ["AnnotationMapper", "ErrorMapper", "ModuleIDNormalizer"]



"""Tests for ModuleIDNormalizer: dot-notation module IDs ↔ OpenAI-compatible names."""

from __future__ import annotations

import re

import pytest

from apcore_mcp.adapters.id_normalizer import ModuleIDNormalizer


class TestModuleIDNormalizer:
    """Test suite for ModuleIDNormalizer."""

    @pytest.fixture
    def normalizer(self) -> ModuleIDNormalizer:
        """Provide a ModuleIDNormalizer instance for tests."""
        return ModuleIDNormalizer()

    def test_normalize_dotted_id(self, normalizer: ModuleIDNormalizer) -> None:
        """Test normalizing a simple dotted module ID."""
        result = normalizer.normalize("image.resize")
        assert result == "image-resize"

    def test_normalize_nested_dots(self, normalizer: ModuleIDNormalizer) -> None:
        """Test normalizing a deeply nested module ID with multiple dots."""
        result = normalizer.normalize("comfyui.image.resize.v2")
        assert result == "comfyui-image-resize-v2"

    def test_normalize_already_valid(self, normalizer: ModuleIDNormalizer) -> None:
        """Test normalizing an ID that's already valid (no dots)."""
        result = normalizer.normalize("simple_module")
        assert result == "simple_module"

    def test_normalize_single_segment(self, normalizer: ModuleIDNormalizer) -> None:
        """Test normalizing a single-segment ID."""
        result = normalizer.normalize("ping")
        assert result == "ping"

    def test_normalize_underscore_preserved(self, normalizer: ModuleIDNormalizer) -> None:
        """Test that underscores are preserved during normalization."""
        result = normalizer.normalize("my_module.sub_module")
        assert result == "my_module-sub_module"

    def test_denormalize_dashed_name(self, normalizer: ModuleIDNormalizer) -> None:
        """Test denormalizing a simple dashed function name."""
        result = normalizer.denormalize("image-resize")
        assert result == "image.resize"

    def test_denormalize_nested(self, normalizer: ModuleIDNormalizer) -> None:
        """Test denormalizing a nested function name with multiple dashes."""
        result = normalizer.denormalize("comfyui-image-resize-v2")
        assert result == "comfyui.image.resize.v2"

    def test_denormalize_no_dashes(self, normalizer: ModuleIDNormalizer) -> None:
        """Test denormalizing a name with no dashes."""
        result = normalizer.denormalize("ping")
        assert result == "ping"

    def test_roundtrip(self, normalizer: ModuleIDNormalizer) -> None:
        """Test that normalize and denormalize are inverse operations."""
        test_ids = [
            "ping",
            "image.resize",
            "comfyui.image.resize.v2",
            "my_module",
            "my_module.sub_module",
            "a.b.c.d.e.f",
        ]

        for module_id in test_ids:
            normalized = normalizer.normalize(module_id)
            denormalized = normalizer.denormalize(normalized)
            assert (
                denormalized == module_id
            ), f"Roundtrip failed for '{module_id}': normalized to '{normalized}', denormalized to '{denormalized}'"

    def test_normalize_empty_string_raises(self, normalizer: ModuleIDNormalizer) -> None:
        """Test normalizing an empty string raises ValueError."""
        with pytest.raises(ValueError, match="Invalid module ID"):
            normalizer.normalize("")

    def test_normalize_invalid_ids_raise(self, normalizer: ModuleIDNormalizer) -> None:
        """Test that invalid module IDs raise ValueError."""
        invalid_ids = [
            "",
            "ABC.DEF.GHI",
            "lower.UPPER.MiXeD",
            "123starts_with_digit",
            ".leading.dot",
            "trailing.dot.",
            "has spaces",
            "has-dashes",
        ]
        for module_id in invalid_ids:
            with pytest.raises(ValueError, match="Invalid module ID"):
                normalizer.normalize(module_id)

    def test_normalize_result_matches_pattern(self, normalizer: ModuleIDNormalizer) -> None:
        """Test that all normalized results match the OpenAI function name pattern."""
        pattern = re.compile(r"^[a-zA-Z0-9_-]*$")

        test_ids = [
            "ping",
            "image.resize",
            "comfyui.image.resize.v2",
            "my_module",
            "my_module.sub_module",
            "a.b.c.d.e.f",
            "test123.module456",
        ]

        for module_id in test_ids:
            normalized = normalizer.normalize(module_id)
            assert pattern.match(
                normalized
            ), f"Normalized result '{normalized}' from '{module_id}' does not match pattern ^[a-zA-Z0-9_-]*$"



"""Tests for OpenAIConverter."""

from __future__ import annotations

import pytest

from apcore_mcp.converters.openai import OpenAIConverter
from tests.conftest import ModuleAnnotations, ModuleDescriptor

# ---------------------------------------------------------------------------
# Lightweight stub for apcore Registry used in converter tests.
# ---------------------------------------------------------------------------


class StubRegistry:
    def __init__(self, descriptors: list[ModuleDescriptor]):
        self._descriptors = {d.module_id: d for d in descriptors}

    def list(self, tags=None, prefix=None):
        ids = list(self._descriptors.keys())
        if prefix is not None:
            ids = [mid for mid in ids if mid.startswith(prefix)]
        if tags is not None:
            tag_set = set(tags)
            ids = [mid for mid in ids if tag_set.issubset(set(self._descriptors[mid].tags))]
        return sorted(ids)

    def get_definition(self, module_id):
        return self._descriptors.get(module_id)


class TestConvertDescriptor:
    """Tests for OpenAIConverter.convert_descriptor."""

    @pytest.fixture
    def converter(self):
        return OpenAIConverter()

    def test_convert_simple_descriptor(self, converter, simple_descriptor):
        """Basic conversion produces a valid OpenAI tool definition."""
        result = converter.convert_descriptor(simple_descriptor)

        assert result["type"] == "function"
        assert "function" in result
        func = result["function"]
        assert func["name"] == "image-resize"
        assert func["description"] == "Resize an image to the specified dimensions"
        assert "parameters" in func

    def test_convert_descriptor_output_format(self, converter, simple_descriptor):
        """Verify exact top-level structure: {"type": "function", "function": {...}}."""
        result = converter.convert_descriptor(simple_descriptor)

        # Only two top-level keys
        assert set(result.keys()) == {"type", "function"}
        assert result["type"] == "function"

        # Function dict must have name, description, parameters
        func = result["function"]
        assert "name" in func
        assert "description" in func
        assert "parameters" in func

        # strict key should NOT be present by default
        assert "strict" not in func

    def test_convert_descriptor_id_normalization(self, converter, simple_descriptor):
        """Dots in module_id are replaced with dashes."""
        result = converter.convert_descriptor(simple_descriptor)
        assert result["function"]["name"] == "image-resize"

    def test_convert_descriptor_id_normalization_deep(self, converter):
        """Multiple dots are all replaced with dashes."""
        descriptor = ModuleDescriptor(
            module_id="comfyui.image.resize.v2",
            description="Deep nested id",
            input_schema={"type": "object", "properties": {}},
            output_schema={},
        )
        result = converter.convert_descriptor(descriptor)
        assert result["function"]["name"] == "comfyui-image-resize-v2"

    def test_convert_descriptor_schema_conversion(self, converter, simple_descriptor):
        """input_schema is properly converted via SchemaConverter."""
        result = converter.convert_descriptor(simple_descriptor)
        params = result["function"]["parameters"]

        # SchemaConverter preserves simple schemas as-is
        assert params["type"] == "object"
        assert "width" in params["properties"]
        assert "height" in params["properties"]
        assert "image_path" in params["properties"]
        assert params["required"] == ["width", "height", "image_path"]

    def test_convert_descriptor_embed_annotations_true(self, converter, simple_descriptor):
        """Annotations suffix is appended to description when embed_annotations=True."""
        result = converter.convert_descriptor(simple_descriptor, embed_annotations=True)
        desc = result["function"]["description"]

        # Should contain the original description
        assert desc.startswith("Resize an image to the specified dimensions")
        # Should contain annotation hints
        assert "[Annotations:" in desc
        assert "idempotent=true" in desc

    def test_convert_descriptor_embed_annotations_false(self, converter, simple_descriptor):
        """No suffix when embed_annotations=False (default)."""
        result = converter.convert_descriptor(simple_descriptor, embed_annotations=False)
        desc = result["function"]["description"]

        assert desc == "Resize an image to the specified dimensions"
        assert "[Annotations:" not in desc

    def test_convert_descriptor_strict_mode(self, converter, simple_descriptor):
        """strict=True adds 'strict': true and sets additionalProperties: false."""
        result = converter.convert_descriptor(simple_descriptor, strict=True)
        func = result["function"]

        assert func["strict"] is True
        params = func["parameters"]
        assert params.get("additionalProperties") is False

    def test_convert_descriptor_strict_mode_all_required(self, converter):
        """In strict mode, all properties appear in the required list (sorted)."""
        descriptor = ModuleDescriptor(
            module_id="test.strict",
            description="Strict test",
            input_schema={
                "type": "object",
                "properties": {
                    "required_field": {"type": "string"},
                    "optional_field": {"type": "integer"},
                },
                "required": ["required_field"],
            },
            output_schema={},
        )
        result = converter.convert_descriptor(descriptor, strict=True)
        params = result["function"]["parameters"]

        # All properties must be in required (sorted alphabetically by to_strict_schema)
        assert params["required"] == ["optional_field", "required_field"]

    def test_convert_descriptor_strict_mode_optional_nullable(self, converter):
        """Optional properties get nullable type in strict mode."""
        descriptor = ModuleDescriptor(
            module_id="test.nullable",
            description="Nullable test",
            input_schema={
                "type": "object",
                "properties": {
                    "required_field": {"type": "string"},
                    "optional_field": {"type": "integer"},
                },
                "required": ["required_field"],
            },
            output_schema={},
        )
        result = converter.convert_descriptor(descriptor, strict=True)
        params = result["function"]["parameters"]

        # required_field should keep its original type
        assert params["properties"]["required_field"]["type"] == "string"

        # optional_field should become nullable
        assert params["properties"]["optional_field"]["type"] == ["integer", "null"]

    def test_convert_descriptor_strict_mode_removes_defaults(self, converter):
        """Strict mode removes default values from properties."""
        descriptor = ModuleDescriptor(
            module_id="test.defaults",
            description="Defaults test",
            input_schema={
                "type": "object",
                "properties": {
                    "name": {"type": "string"},
                    "count": {"type": "integer", "default": 10},
                },
                "required": ["name"],
            },
            output_schema={},
        )
        result = converter.convert_descriptor(descriptor, strict=True)
        params = result["function"]["parameters"]

        # default should be removed
        assert "default" not in params["properties"]["count"]

    def test_convert_descriptor_strict_mode_nested_objects(self, converter):
        """Strict mode recurses into nested objects."""
        descriptor = ModuleDescriptor(
            module_id="test.nested_strict",
            description="Nested strict test",
            input_schema={
                "type": "object",
                "properties": {
                    "config": {
                        "type": "object",
                        "properties": {
                            "key": {"type": "string"},
                            "value": {"type": "string"},
                        },
                        "required": ["key"],
                    },
                },
                "required": ["config"],
            },
            output_schema={},
        )
        result = converter.convert_descriptor(descriptor, strict=True)
        nested = result["function"]["parameters"]["properties"]["config"]

        assert nested.get("additionalProperties") is False
        assert nested["required"] == ["key", "value"]  # sorted by to_strict_schema
        # value was optional, so it should be nullable
        assert nested["properties"]["value"]["type"] == ["string", "null"]

    def test_convert_descriptor_no_annotations(self, converter, no_annotations_descriptor):
        """Works when annotations is None."""
        result = converter.convert_descriptor(no_annotations_descriptor)

        assert result["type"] == "function"
        func = result["function"]
        assert func["name"] == "text-echo"
        assert func["description"] == "Echo the input text"

    def test_convert_descriptor_no_annotations_embed(self, converter, no_annotations_descriptor):
        """embed_annotations=True with None annotations produces no suffix."""
        result = converter.convert_descriptor(
            no_annotations_descriptor,
            embed_annotations=True,
        )
        desc = result["function"]["description"]
        # AnnotationMapper.to_description_suffix returns "" for None
        assert desc == "Echo the input text"

    def test_convert_descriptor_empty_schema(self, converter, empty_schema_descriptor):
        """Handles empty input_schema (SchemaConverter fills in defaults)."""
        result = converter.convert_descriptor(empty_schema_descriptor)
        params = result["function"]["parameters"]

        # SchemaConverter converts {} to {"type": "object", "properties": {}}
        assert params == {"type": "object", "properties": {}}

    def test_convert_descriptor_destructive(self, converter, destructive_descriptor):
        """Destructive descriptor is correctly converted with annotation embed."""
        result = converter.convert_descriptor(destructive_descriptor, embed_annotations=True)
        desc = result["function"]["description"]

        assert "destructive=true" in desc
        assert "requires_approval=true" in desc


class TestConvertRegistry:
    """Tests for OpenAIConverter.convert_registry."""

    @pytest.fixture
    def converter(self):
        return OpenAIConverter()

    def test_convert_registry_multiple_modules(self, converter):
        """Converts all modules in a registry."""
        registry = StubRegistry(
            [
                ModuleDescriptor(
                    module_id="mod.a",
                    description="Module A",
                    input_schema={
                        "type": "object",
                        "properties": {"x": {"type": "string"}},
                    },
                    output_schema={},
                    tags=["core"],
                ),
                ModuleDescriptor(
                    module_id="mod.b",
                    description="Module B",
                    input_schema={
                        "type": "object",
                        "properties": {"y": {"type": "integer"}},
                    },
                    output_schema={},
                    tags=["core"],
                ),
            ]
        )
        results = converter.convert_registry(registry)

        assert len(results) == 2
        names = {r["function"]["name"] for r in results}
        assert names == {"mod-a", "mod-b"}

    def test_convert_registry_empty(self, converter):
        """Returns empty list for empty registry."""
        registry = StubRegistry([])
        results = converter.convert_registry(registry)

        assert results == []

    def test_convert_registry_tag_filter(self, converter):
        """Only includes modules matching tags."""
        registry = StubRegistry(
            [
                ModuleDescriptor(
                    module_id="image.resize",
                    description="Resize",
                    input_schema={"type": "object", "properties": {}},
                    output_schema={},
                    tags=["image", "transform"],
                ),
                ModuleDescriptor(
                    module_id="text.echo",
                    description="Echo",
                    input_schema={"type": "object", "properties": {}},
                    output_schema={},
                    tags=["text"],
                ),
            ]
        )
        results = converter.convert_registry(registry, tags=["image"])

        assert len(results) == 1
        assert results[0]["function"]["name"] == "image-resize"

    def test_convert_registry_prefix_filter(self, converter):
        """Only includes modules matching prefix."""
        registry = StubRegistry(
            [
                ModuleDescriptor(
                    module_id="image.resize",
                    description="Resize",
                    input_schema={"type": "object", "properties": {}},
                    output_schema={},
                    tags=[],
                ),
                ModuleDescriptor(
                    module_id="text.echo",
                    description="Echo",
                    input_schema={"type": "object", "properties": {}},
                    output_schema={},
                    tags=[],
                ),
            ]
        )
        results = converter.convert_registry(registry, prefix="text")

        assert len(results) == 1
        assert results[0]["function"]["name"] == "text-echo"

    def test_convert_registry_skip_none_definition(self, converter):
        """Skips when get_definition returns None (race condition)."""

        class HoleyRegistry:
            """Registry that returns None for some module IDs."""

            def list(self, tags=None, prefix=None):
                return ["mod.exists", "mod.gone"]

            def get_definition(self, module_id):
                if module_id == "mod.exists":
                    return ModuleDescriptor(
                        module_id="mod.exists",
                        description="Exists",
                        input_schema={"type": "object", "properties": {}},
                        output_schema={},
                    )
                return None  # Simulate race condition

        registry = HoleyRegistry()
        results = converter.convert_registry(registry)

        assert len(results) == 1
        assert results[0]["function"]["name"] == "mod-exists"

    def test_convert_registry_passes_embed_and_strict(self, converter):
        """embed_annotations and strict are forwarded to convert_descriptor."""
        registry = StubRegistry(
            [
                ModuleDescriptor(
                    module_id="mod.a",
                    description="Module A",
                    input_schema={
                        "type": "object",
                        "properties": {"x": {"type": "string"}},
                        "required": ["x"],
                    },
                    output_schema={},
                    annotations=ModuleAnnotations(destructive=True),
                ),
            ]
        )
        results = converter.convert_registry(
            registry,
            embed_annotations=True,
            strict=True,
        )

        assert len(results) == 1
        func = results[0]["function"]
        assert func["strict"] is True
        assert "destructive=true" in func["description"]
        assert func["parameters"].get("additionalProperties") is False

    def test_convert_registry_preserves_order(self, converter):
        """Results follow the order returned by registry.list()."""
        registry = StubRegistry(
            [
                ModuleDescriptor(
                    module_id="b.mod",
                    description="B",
                    input_schema={"type": "object", "properties": {}},
                    output_schema={},
                ),
                ModuleDescriptor(
                    module_id="a.mod",
                    description="A",
                    input_schema={"type": "object", "properties": {}},
                    output_schema={},
                ),
            ]
        )
        results = converter.convert_registry(registry)
        names = [r["function"]["name"] for r in results]
        # StubRegistry.list() returns sorted, so a.mod comes before b.mod
        assert names == ["a-mod", "b-mod"]


class TestStrictModeEdgeCases:
    """Tests for strict mode edge cases in _apply_strict_recursive."""

    @pytest.fixture
    def converter(self):
        return OpenAIConverter()

    def test_strict_mode_list_type_already_has_null(self, converter):
        """Strict mode does not duplicate 'null' when type is already a list with null."""
        descriptor = ModuleDescriptor(
            module_id="test.list_null",
            description="Test",
            input_schema={
                "type": "object",
                "properties": {
                    "req_field": {"type": "string"},
                    "opt_field": {"type": ["string", "null"]},
                },
                "required": ["req_field"],
            },
            output_schema={},
        )

        result = converter.convert_descriptor(descriptor, strict=True)
        params = result["function"]["parameters"]
        # opt_field already has null in type list, should not duplicate
        assert params["properties"]["opt_field"]["type"] == ["string", "null"]

    def test_strict_mode_list_type_without_null(self, converter):
        """Strict mode appends null to list type that doesn't have it."""
        descriptor = ModuleDescriptor(
            module_id="test.list_no_null",
            description="Test",
            input_schema={
                "type": "object",
                "properties": {
                    "req_field": {"type": "string"},
                    "opt_field": {"type": ["string", "integer"]},
                },
                "required": ["req_field"],
            },
            output_schema={},
        )

        result = converter.convert_descriptor(descriptor, strict=True)
        params = result["function"]["parameters"]
        assert params["properties"]["opt_field"]["type"] == [
            "string",
            "integer",
            "null",
        ]

    def test_strict_mode_array_items_recursion(self, converter):
        """Strict mode recurses into array items."""
        descriptor = ModuleDescriptor(
            module_id="test.array_recurse",
            description="Test",
            input_schema={
                "type": "object",
                "properties": {
                    "items": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "name": {"type": "string"},
                                "value": {"type": "integer", "default": 0},
                            },
                            "required": ["name"],
                        },
                    },
                },
                "required": ["items"],
            },
            output_schema={},
        )

        result = converter.convert_descriptor(descriptor, strict=True)
        params = result["function"]["parameters"]
        item_schema = params["properties"]["items"]["items"]
        # Items should have strict mode applied too
        assert item_schema["additionalProperties"] is False
        assert item_schema["required"] == [
            "name",
            "value",
        ]  # sorted by to_strict_schema
        # Default should be removed
        assert "default" not in item_schema["properties"]["value"]

    def test_strict_mode_strips_x_extensions(self, converter):
        """to_strict_schema() strips x-* extension fields."""
        descriptor = ModuleDescriptor(
            module_id="test.extensions",
            description="Test",
            input_schema={
                "type": "object",
                "properties": {
                    "name": {
                        "type": "string",
                        "x-llm-description": "A custom hint",
                        "x-sensitive": True,
                    },
                },
                "required": ["name"],
            },
            output_schema={},
        )

        result = converter.convert_descriptor(descriptor, strict=True)
        params = result["function"]["parameters"]
        # x-* fields should be stripped
        assert "x-llm-description" not in params["properties"]["name"]
        assert "x-sensitive" not in params["properties"]["name"]

    def test_strict_mode_promotes_x_llm_description(self, converter):
        """x-llm-description is promoted to description before stripping."""
        descriptor = ModuleDescriptor(
            module_id="test.llm_desc",
            description="Test",
            input_schema={
                "type": "object",
                "properties": {
                    "query": {
                        "type": "string",
                        "description": "Original description",
                        "x-llm-description": "LLM-optimized description",
                    },
                },
                "required": ["query"],
            },
            output_schema={},
        )

        result = converter.convert_descriptor(descriptor, strict=True)
        params = result["function"]["parameters"]
        # x-llm-description should have been promoted to description
        assert params["properties"]["query"]["description"] == "LLM-optimized description"
        # x-llm-description key itself should be stripped
        assert "x-llm-description" not in params["properties"]["query"]



"""ExecutionRouter: route MCP tool calls -> apcore Executor pipeline."""

from __future__ import annotations

import inspect
import json
import logging
from collections.abc import Callable, Coroutine
from typing import Any

from apcore import Context

from apcore_mcp.adapters.errors import ErrorMapper
from apcore_mcp.helpers import MCP_ELICIT_KEY, MCP_PROGRESS_KEY

logger = logging.getLogger(__name__)


class ExecutionRouter:
    """Routes MCP tool calls through the apcore Executor pipeline.

    The router sits between the MCP server's call_tool handler and the
    apcore Executor.  It delegates to ``executor.call_async()`` and
    converts the result (or any exception) into a ``(content, is_error)``
    tuple that the MCP factory can pass directly to ``CallToolResult``.

    When the executor also exposes an async ``stream()`` method **and**
    the caller provides a ``progress_token`` + ``send_notification``
    callback via the *extra* dict, the router iterates the async
    generator and forwards each chunk as a ``notifications/progress``
    message, accumulating chunks via shallow merge.

    Args:
        executor: An apcore Executor instance (duck-typed -- must expose
            an async ``call_async(module_id, inputs)`` method and
            optionally an async ``stream(module_id, inputs)`` generator).
    """

    def __init__(self, executor: Any, *, validate_inputs: bool = False) -> None:
        self._executor = executor
        self._error_mapper = ErrorMapper()
        self._validate_inputs = validate_inputs

        # Cache whether executor methods accept a context parameter,
        # so we avoid a broad TypeError catch on every call.
        self._call_async_accepts_context = self._check_accepts_context(executor.call_async)
        self._stream_accepts_context = self._check_accepts_context(getattr(executor, "stream", None))

    @staticmethod
    def _check_accepts_context(method: Any) -> bool:
        """Return True if *method* accepts at least 3 positional parameters
        (excluding ``self``), i.e. (tool_name, arguments, context)."""
        if method is None:
            return False
        try:
            sig = inspect.signature(method)
            return len(sig.parameters) >= 3
        except (ValueError, TypeError):
            return True  # assume yes if we cannot inspect

    async def handle_call(
        self,
        tool_name: str,
        arguments: dict[str, Any],
        extra: dict[str, Any] | None = None,
    ) -> tuple[list[dict[str, str]], bool]:
        """Execute a tool call through the Executor pipeline.

        Args:
            tool_name: The MCP tool name (already denormalized to apcore
                module ID by the caller, or passed through as-is).
            arguments: The tool call arguments dict.
            extra: Optional dict with ``progress_token``,
                ``send_notification``, and ``session`` for streaming
                and elicitation support.

        Returns:
            A ``(content, is_error)`` tuple where *content* is a list of
            ``TextContent``-compatible dicts and *is_error* signals
            whether the result represents an error.
        """
        logger.debug("Executing tool call: %s", tool_name)

        # Extract streaming helpers from extra
        progress_token: str | int | None = None
        send_notification: Callable[[dict[str, Any]], Coroutine[Any, Any, None]] | None = None
        session: Any = None
        if extra is not None:
            progress_token = extra.get("progress_token")
            send_notification = extra.get("send_notification")
            session = extra.get("session")

        # ── Build context with MCP callbacks ─────────────────────────────
        context_data: dict[str, Any] = {}

        # Inject progress callback if progress_token + send_notification available
        if progress_token is not None and send_notification is not None:
            _pt = progress_token
            _sn = send_notification

            async def _progress_callback(
                progress: float,
                total: float | None = None,
                message: str | None = None,
            ) -> None:
                notification: dict[str, Any] = {
                    "method": "notifications/progress",
                    "params": {
                        "progressToken": _pt,
                        "progress": progress,
                        "total": total if total is not None else 0,
                    },
                }
                if message is not None:
                    notification["params"]["message"] = message
                await _sn(notification)

            context_data[MCP_PROGRESS_KEY] = _progress_callback

        # Inject elicitation callback if session available
        if session is not None:
            _session = session

            async def _elicit_callback(
                message: str,
                requested_schema: dict[str, Any] | None = None,
            ) -> dict[str, Any] | None:
                try:
                    result = await _session.elicit_form(
                        message=message,
                        requestedSchema=requested_schema or {},
                    )
                    return {
                        "action": result.action,
                        "content": result.content,
                    }
                except Exception:
                    logger.debug("Elicitation request failed", exc_info=True)
                    return None

            context_data[MCP_ELICIT_KEY] = _elicit_callback

        context = Context.create(data=context_data)

        # Pre-execution validation
        if self._validate_inputs:
            try:
                validation = self._executor.validate(tool_name, arguments)
                if not validation.valid:
                    detail = "; ".join(
                        f"{e.get('field', '?')}: {e.get('message', 'invalid')}" for e in validation.errors
                    )
                    return (
                        [{"type": "text", "text": f"Validation failed: {detail}"}],
                        True,
                    )
            except AttributeError:
                pass  # executor lacks validate() — skip
            except Exception as error:
                logger.debug("validate_inputs error for %s: %s", tool_name, error)
                error_info = self._error_mapper.to_mcp_error(error)
                return ([{"type": "text", "text": error_info["message"]}], True)

        # Streaming path: executor has stream() AND we have both helpers
        can_stream = hasattr(self._executor, "stream") and progress_token is not None and send_notification is not None

        if can_stream:
            return await self._handle_stream(
                tool_name,
                arguments,
                progress_token,  # type: ignore[arg-type]
                send_notification,  # type: ignore[arg-type]
                context=context,
            )

        # Non-streaming path
        return await self._handle_call_async(tool_name, arguments, context=context)

    async def _handle_call_async(
        self,
        tool_name: str,
        arguments: dict[str, Any],
        context: Any | None = None,
    ) -> tuple[list[dict[str, str]], bool]:
        """Non-streaming execution via executor.call_async()."""
        try:
            if self._call_async_accepts_context:
                result = await self._executor.call_async(tool_name, arguments, context)
            else:
                result = await self._executor.call_async(tool_name, arguments)
            json_output = json.dumps(result, default=str)
            content: list[dict[str, str]] = [{"type": "text", "text": json_output}]
            if context is not None:
                content.append(
                    {
                        "type": "text",
                        "text": json.dumps({"_trace_id": context.trace_id}),
                    }
                )
            return (content, False)
        except Exception as error:
            logger.error("handle_call error for %s: %s", tool_name, error)
            error_info = self._error_mapper.to_mcp_error(error)
            return ([{"type": "text", "text": error_info["message"]}], True)

    async def _handle_stream(
        self,
        tool_name: str,
        arguments: dict[str, Any],
        progress_token: str | int,
        send_notification: Callable[[dict[str, Any]], Coroutine[Any, Any, None]],
        context: Any | None = None,
    ) -> tuple[list[dict[str, str]], bool]:
        """Streaming execution via executor.stream().

        Iterates the async generator, sends each chunk as a
        ``notifications/progress`` message, accumulates via shallow
        merge, and returns the final accumulated result.
        """
        accumulated: dict[str, Any] = {}
        chunk_index = 0

        try:
            if self._stream_accepts_context:
                stream_iter = self._executor.stream(tool_name, arguments, context)
            else:
                stream_iter = self._executor.stream(tool_name, arguments)

            async for chunk in stream_iter:
                # Send progress notification for this chunk
                notification: dict[str, Any] = {
                    "method": "notifications/progress",
                    "params": {
                        "progressToken": progress_token,
                        "progress": chunk_index + 1,
                        "total": None,
                        "message": json.dumps(chunk, default=str),
                    },
                }
                await send_notification(notification)

                # Shallow merge into accumulated result
                accumulated = {**accumulated, **chunk}
                chunk_index += 1

            json_output = json.dumps(accumulated, default=str)
            content: list[dict[str, str]] = [{"type": "text", "text": json_output}]
            if context is not None:
                content.append(
                    {
                        "type": "text",
                        "text": json.dumps({"_trace_id": context.trace_id}),
                    }
                )
            return (content, False)
        except Exception as error:
            logger.error("handle_call stream error for %s: %s", tool_name, error)
            error_info = self._error_mapper.to_mcp_error(error)
            return ([{"type": "text", "text": error_info["message"]}], True)
